--[[

States:
-3	Just blackness
-2  tostate
-1	Display error (expected: errormsg)
0	Temp main menu (enter state)
1	The editor (will expect things to have been loaded!)
2	Syntax highlighting test
3	Scripting editor
4	Some XML testing
5	Filesystem testing
6	Simple listing of all files in the levels folder, and load a level from here
7	Display all sprites from sprites.png where you can get the number of the sprite you're hovering over
8	Just save by going to this state and typing in a name
9	Dialog test, and right click menu test
10	List of scripts, and enter one to load
11	Search
12	Map
13	Options screen
14	Sort of entity picker proto
15	Help
16	Scroll bar test
17	folderopendialog utility

Debug keys:
F12: change state
F11: dump all variables in console
' (in editor): display all tilesets and colors in the console
\ (in editor): tileset creator/compiler [NOW CHANGED TO CTRL+\]
/ (in editor): instead of displaying entities as themselves, display them as an E-block
PgDn: change FPS cap

]]

function love.load()
	require("loadconfig")
	require("lang/English") -- just as a fallback. Doesn't work for the most part.
	if s.language ~= "English" then
		require("lang/" .. s.language)
	end
	require("const")
	require("func")
	require("roomfunc")
	require("vvvvvvxml")
	require("dialog")
	
	if s.pscale ~= 1 then
		za,zb,zc = love.window.getMode()
		love.window.setMode(za*s.pscale,zb*s.pscale,zc)
		
		require("scaling")
	end

	if love.window == nil then
		-- Our version of löve is too old, display error message. (update; this specific code never triggers anymore because love 0.8 and lower is handled in main.lua, should remove this)
		print("Outdated love2d version!")
		errormsg = L.OUTDATEDLOVE
		state = -1
	else
		-- Are we using 0.10.x?
		if love.getVersion ~= nil then
			local major, minor, revision, codename = love.getVersion()
			
			if minor >= 10 then
				require("love10compat")
			end
		end
		
		cons("love.load() reached")
		
		state = -2; oldstate = "none"
		debug = 0
		input = ""; takinginput = false
		__ = "_"
		input_r = ""
		cursorflashtime = 0
		multiinput = {}; currentmultiinput = 0
		bypassdialog = false
		
		mousepressed = false -- for some things
		
		temporaryroomnametimer = 0
		
		-- Dialogs. To not have an active window, set DIAwindowani to 16. To show one, just call dialog.new(message, title, showbar, canclose, questionid). -15 ~ -1 opening, 0 is normal, 1 ~ 15 closing
		dialog.init()
		nodialog = true
		
		tilesets = {}
		
		-- Load a couple of images
		cursorimg = {}
		cursorimg[0] = love.graphics.newImage("cursor/cursor.png")
		cursorimg[1] = love.graphics.newImage("cursor/cursor1.png")
		cursorimg[2] = love.graphics.newImage("cursor/cursor2.png")
		cursorimg[3] = love.graphics.newImage("cursor/cursor3.png")
		cursorimg[4] = love.graphics.newImage("cursor/cursor4.png")
		cursorimg[5] = love.graphics.newImage("cursor/entity.png")
		cursorimg[6] = love.graphics.newImage("cursor/specialentity.png")
		
		selectedtoolborder = love.graphics.newImage("selectedtool.png")
		unselectedtoolborder = love.graphics.newImage("unselectedtool.png")
		
		savebtn = love.graphics.newImage("save.png")
		loadbtn = love.graphics.newImage("load.png")
		newbtn = love.graphics.newImage("new.png")
		helpbtn = love.graphics.newImage("help.png")
		
		checkon = love.graphics.newImage("checkon.png")
		checkoff = love.graphics.newImage("checkoff.png")
		
		bggrid = love.graphics.newImage("bggrid.png")
		
		toolimg = {}
		toolimgicon = {}
		for t = 1, 17 do
			toolimg[t] = love.graphics.newImage("tools/" .. t .. ".png")
			toolimgicon[t] = love.image.newImageData("tools2/on/" .. t .. ".png")
		end
		
		-- But we also have subtools!
		subtoolimgs = {}
		subtoolimgs[1] = {st("1_1"), st("1_2"), st("1_3"), st("1_4"), st("1_5"), st("1_6"), st("1_7"), st("1_8")}
		subtoolimgs[2] = {st("1_1"), st("1_2"), st("1_3"), st("1_4"), st("1_5"), st("1_6"), st("1_7"), st("1_8")}
		subtoolimgs[3] = {st("3_1"), st("3_2"), st("3_3"), st("3_4"), }-- st("3_5"), st("3_6"), st("3_7"), st("3_8")}
		subtoolimgs[4] = {}
		subtoolimgs[5] = {st("5_1"), st("5_2")}
		subtoolimgs[6] = {}
		subtoolimgs[7] = {st("7_1"), st("7_2"), st("7_3"), st("7_4")}
		subtoolimgs[8] = {st("8_2"), st("8_1"), st("8_3"), st("8_4")}
		subtoolimgs[9] = {st("9_2"), st("9_1"), st("9_3"), st("9_4")}
		subtoolimgs[10] = {st("10_1"), st("10_2")}
		subtoolimgs[11] = {}
		subtoolimgs[12] = {}
		subtoolimgs[13] = {}
		subtoolimgs[14] = {st("14_1"), st("14_2")}
		subtoolimgs[15] = {}
		subtoolimgs[16] = {st("16_1"), st("16_2"), st("16_3"), st("16_4"), st("16_5"), st("16_6")}
		subtoolimgs[17] = {st("17_1"), st("17_2")}
		
		
		scrollup = love.graphics.newImage("scrollup.png")
		scrolldn = love.graphics.newImage("scrolldn.png")
		
		sideimg = love.graphics.newImage("sides.png");             smallsideimg = love.graphics.newImage("smallsides.png")
		
		sideline = {};                                             smallsideline = {}
		sideline[1] = love.graphics.newQuad(0, 0, 8, 8, 32, 8);    smallsideline[1] = love.graphics.newQuad(0, 0, 16, 16, 64, 16)
		sideline[2] = love.graphics.newQuad(8, 0, 8, 8, 32, 8);    smallsideline[2] = love.graphics.newQuad(16, 0, 16, 16, 64, 16)
		sideline[3] = love.graphics.newQuad(16, 0, 8, 8, 32, 8);   smallsideline[3] = love.graphics.newQuad(32, 0, 16, 16, 64, 16)
		sideline[4] = love.graphics.newQuad(24, 0, 8, 8, 32, 8);   smallsideline[4] = love.graphics.newQuad(48, 0, 16, 16, 64, 16)
		
		platformimg = love.graphics.newImage("platform.png")
		platformpart =
			{
			love.graphics.newQuad(0, 0, 8, 8, 24, 8),
			love.graphics.newQuad(8, 0, 8, 8, 24, 8),
			love.graphics.newQuad(16, 0, 8, 8, 24, 8)
			}
			
		-- The help has images too, but they shouldn't be loaded repetitively!
		helpimages = {}
		
		-- Load the default tilesets - Will be done later when we have the custom level folder
		--loadtileset("tiles.png")
		--loadtileset("tiles2.png")
		-- loadtileset("tiles3.png") oops oversight
		
		-- Load the spriteset - This later too
		-- loadsprites("sprites.png", 64)
		--loadsprites("sprites.png", 32)
		
		-- eeeeeeeeee
		love.keyboard.setKeyRepeat(true)
		thingk()
		
		if love.system.getOS() == "OS X" then
			-- Cmd
			ctrl = "gui"
			dirsep = "/"
			require("filefunc_mac")
		elseif love.system.getOS() == "Windows" then
			-- Ctrl
			ctrl = "ctrl"
			dirsep = "\\"
			require("filefunc_win")
		elseif love.system.getOS() == "Linux" then
			-- Ctrl
			ctrl = "ctrl"
			dirsep = "/"
			require("filefunc_lin")
			--require("filefunc_luv")
			--dialog.new("Full support for Linux will be added really soon!", "", 1, 1, 0)
		else
			-- This OS is unknown, so I suppose we will have to fall back on functions in love.filesystem.
			ctrl = "ctrl"
			dirsep = "/"
			require("filefunc_luv")
			dialog.new(langkeys(L.OSNOTRECOGNIZED, {anythingbutnil(love.system.getOS()), love.filesystem.getSaveDirectory()}), "", 1, 1, 0)
		end
				
		if love.getVersion ~= nil then
			function openurl(url)
				love.system.openURL(url)
			end
		end
		
		--if not love.filesystem.exists("temp") then
			--love.filesystem.createDirectory("temp")
		--end
		if not love.filesystem.exists("maps") then
			love.filesystem.createDirectory("maps")
		end
		
		updatecheckthread = love.thread.newThread("updatecheck.lua")
		
		verchannel = love.thread.getChannel("version")
		updatecheckthread:start(checkver)
	end
end

function love.draw()
	if s.pscale ~= 1 then
		love.graphics.push()
		love.graphics.scale(s.pscale,s.pscale)
	end

	if state == -3 then
		-- Do-nothing state
	elseif state == -2 then
		tostate(6)
	elseif state == -1 then
		love.graphics.printf(L.FATALERROR .. anythingbutnil(errormsg) .. "\n\n" .. L.FATALEND, 10, 10, love.graphics.getWidth()-20, "left")
	elseif state == 0 then
		love.graphics.print("Placeholder main menu. Enter state: " .. input .. __, 10, 10)
		startinputonce()
	elseif state == 1 then
		-- Ok first of all are we clicking?
		
		-- Fix something first...
		if vedmetadata == nil then
			cons("vedmetadata == nil! Setting to false.")
			vedmetadata = false
		end
		
		if nodialog and (love.mouse.isDown("l") or love.mouse.isDown("r")) and mouseon(64+64, 0, 639, 480) then
			local atx = math.floor((love.mouse.getX()-64-64) / 16)
			local aty = math.floor((love.mouse.getY()) / 16)
			
			if tilespicker then
				if false and tilescreator then
					-- This is the one where you have to click 30 tiles individually.
					if not mousepressed then
						if creatorstep == 1 then
							cons("Wall tile " .. creatorsubstep .. ": " .. (aty*40)+(atx+1)-1)
							
							tilesetblocks.creator.colors.creator.blocks[creatorsubstep] = (aty*40)+(atx+1)-1
							cb[creatorsubstep] = (aty*40)+(atx+1)-1
						elseif creatorstep == 2 then
							cons("Background tile " .. creatorsubstep .. ": " .. (aty*40)+(atx+1)-1)
							
							tilesetblocks.creator.colors.creator.background[creatorsubstep] = (aty*40)+(atx+1)-1
							ca[creatorsubstep] = (aty*40)+(atx+1)-1
						elseif creatorstep == 3 then
							cons("Spike tile " .. creatorsubstep .. ": " .. (aty*40)+(atx+1)-1)
							
							tilesetblocks.creator.colors.creator.spikes[creatorsubstep] = (aty*40)+(atx+1)-1
							cs[creatorsubstep] = (aty*40)+(atx+1)-1
						end
						
						if creatorsubstep == 30 then
							creatorstep = creatorstep + 1
							cons("Next: step " .. creatorstep)
							selectedtool = creatorstep
							
							creatorsubstep = 1
						else
							creatorsubstep = creatorsubstep + 1
						end
						
						if creatorstep == 4 then
							spitoutarrays()
							cons("*** ARRAYS COPIED TO CLIPBOARD ***")
							tilescreator = false
						end
						
						mousepressed = true
					end
				elseif tilescreator then
					-- Faster one.
					if not mousepressed then
						if creatorstep == 1 then
							cons("Wall tile " .. creatorsubstep .. ": " .. (aty*40)+(atx+1)-1)
							
							--tilesetblocks.creator.colors.creator.blocks[creatorsubstep] = (aty*40)+(atx+1)-1
							
							for yplus = 0, 4 do
								for xplus = 0, 2 do
									if (aty*40)+(atx+1)-1 == 0 then
										cb[creatorsubstep+(6*yplus)+xplus] = 0
										tilesetblocks.creator.colors.creator.blocks[creatorsubstep+(6*yplus)+xplus] = 0
									else
										cb[creatorsubstep+(6*yplus)+xplus] = (aty*40)+(atx+1)-1 + (40*yplus) + xplus
										tilesetblocks.creator.colors.creator.blocks[creatorsubstep+(6*yplus)+xplus] = (aty*40)+(atx+1)-1 + (40*yplus) + xplus
									end
								end
							end
						elseif creatorstep == 2 then
							cons("Background tile " .. creatorsubstep .. ": " .. (aty*40)+(atx+1)-1)
							
							--tilesetblocks.creator.colors.creator.background[creatorsubstep] = (aty*40)+(atx+1)-1
							
							for yplus = 0, 4 do
								for xplus = 0, 2 do
									if (aty*40)+(atx+1)-1 == 0 then
										ca[creatorsubstep+(6*yplus)+xplus] = 0
										tilesetblocks.creator.colors.creator.blocks[creatorsubstep+(6*yplus)+xplus] = 0
									else
										ca[creatorsubstep+(6*yplus)+xplus] = (aty*40)+(atx+1)-1 + (40*yplus) + xplus
										tilesetblocks.creator.colors.creator.blocks[creatorsubstep+(6*yplus)+xplus] = (aty*40)+(atx+1)-1 + (40*yplus) + xplus
									end
								end
							end
						elseif creatorstep == 3 then
							cons("Spike tile " .. creatorsubstep .. ": " .. (aty*40)+(atx+1)-1)
							
							--tilesetblocks.creator.colors.creator.spikes[creatorsubstep] = (aty*40)+(atx+1)-1
							
							for yplus = 0, 4 do
								for xplus = 0, 2 do
									if (aty*40)+(atx+1)-1 == 0 then
										cs[creatorsubstep+(6*yplus)+xplus] = 0
										tilesetblocks.creator.colors.creator.blocks[creatorsubstep+(6*yplus)+xplus] = 0
									else
										cs[creatorsubstep+(6*yplus)+xplus] = (aty*40)+(atx+1)-1 + (40*yplus) + xplus
										tilesetblocks.creator.colors.creator.blocks[creatorsubstep+(6*yplus)+xplus] = (aty*40)+(atx+1)-1 + (40*yplus) + xplus
									end
								end
							end
						end
						
						if creatorsubstep == 4 then
							creatorstep = creatorstep + 1
							cons("Next: step " .. creatorstep)
							selectedtool = creatorstep
							
							creatorsubstep = 1
						else
							creatorsubstep = creatorsubstep + 3
						end
						
						if creatorstep == 4 then
							spitoutarrays()
							cons("*** ARRAYS COPIED TO CLIPBOARD ***")
							tilescreator = false
						end
						
						mousepressed = true
					end

				elseif levelmetadata[(roomy)*20 + (roomx+1)].directmode == 1 then
					cons("Tile selected: " .. (aty*40)+(atx+1)-1)
				
					selectedtile = (aty*40)+(atx+1)-1
				end
			elseif selectedtool == 1 or selectedtool == 2 then
				cons("Tile clicked: " .. atx .. " " .. aty .. ", set to " .. selectedtile .. ", subtool " .. selectedsubtool[selectedtool])
				
				if love.mouse.isDown("r") then
					useselectedtile = 0
				else
					if levelmetadata[(roomy)*20 + (roomx+1)].directmode == 0 then
						-- Auto mode
						useselectedtile = selectedtool
					else
						useselectedtile = selectedtile
					end
				end
				
				if selectedsubtool[selectedtool] == 1 then
					-- 1x1
					roomdata[roomy][roomx][(aty*40)+(atx+1)] = useselectedtile
				elseif selectedsubtool[selectedtool] == 2 then
					-- 3x3
					for sty = (aty-1), (aty+1) do
						for stx = (atx-1), (atx+1) do
							--if roomdata[roomy][roomx][(sty*40)+(stx+1)] ~= nil then
							if stx >= 0 and stx <= 39 and sty >= 0 and sty <= 29 then
								roomdata[roomy][roomx][(sty*40)+(stx+1)] = useselectedtile
							end
						end
					end
				elseif selectedsubtool[selectedtool] == 3 then
					-- 5x5
					for sty = (aty-2), (aty+2) do
						for stx = (atx-2), (atx+2) do
							if stx >= 0 and stx <= 39 and sty >= 0 and sty <= 29 then
								roomdata[roomy][roomx][(sty*40)+(stx+1)] = useselectedtile
							end
						end
					end
				elseif selectedsubtool[selectedtool] == 4 then
					-- 7x7
					for sty = (aty-3), (aty+3) do
						for stx = (atx-3), (atx+3) do
							if stx >= 0 and stx <= 39 and sty >= 0 and sty <= 29 then
								roomdata[roomy][roomx][(sty*40)+(stx+1)] = useselectedtile
							end
						end
					end
				elseif selectedsubtool[selectedtool] == 5 then
					-- 9x9
					for sty = (aty-4), (aty+4) do
						for stx = (atx-4), (atx+4) do
							if stx >= 0 and stx <= 39 and sty >= 0 and sty <= 29 then
								roomdata[roomy][roomx][(sty*40)+(stx+1)] = useselectedtile
							end
						end
					end
				elseif selectedsubtool[selectedtool] == 6 then
					-- horizontal fill
					if minsmear == -1 and maxsmear == -1 then
						minsmear = aty; maxsmear = aty
						for stx = 0, 39 do
							roomdata[roomy][roomx][(aty*40)+(stx+1)] = useselectedtile
						end
					end
					
					if aty < minsmear or aty > maxsmear then
						for sty = math.min(aty, minsmear), math.max(aty, maxsmear) do
							for stx = 0, 39 do
								roomdata[roomy][roomx][(sty*40)+(stx+1)] = useselectedtile
							end
						end
					end
					
					if aty < minsmear then
						minsmear = aty
					elseif aty > maxsmear then
						maxsmear = aty
					end
				elseif selectedsubtool[selectedtool] == 7 then
					-- vertical fill
					if minsmear == -1 and maxsmear == -1 then
						minsmear = atx; maxsmear = atx
						for sty = 0, 29 do
							roomdata[roomy][roomx][(sty*40)+(atx+1)] = useselectedtile
						end
					end
					
					if atx < minsmear or atx > maxsmear then
						for stx = math.min(atx, minsmear), math.max(atx, maxsmear) do
							for sty = 0, 29 do
								roomdata[roomy][roomx][(sty*40)+(stx+1)] = useselectedtile
							end
						end
					end
					
					if atx < minsmear then
						minsmear = atx
					elseif atx > maxsmear then
						maxsmear = atx
					end
				elseif not mousepressed and selectedsubtool[selectedtool] == 8 then
					-- room fill
					for sty = 0, 29 do
						for stx = 0, 39 do
							roomdata[roomy][roomx][(sty*40)+(stx+1)] = useselectedtile
						end
					end
					
					mousepressed = true
				elseif selectedsubtool[selectedtool] == 9 then
					-- to out
					-- rot
					
					for rot = 0, toout-1 do
						local tooutnow = toout - rot
						roomdata[roomy][roomx][((aty-tooutnow)*40)+((atx+rot)+1)] = useselectedtile -- top to right
						roomdata[roomy][roomx][((aty+rot)*40)+((atx+tooutnow)+1)] = useselectedtile -- right to bottom
						roomdata[roomy][roomx][((aty+tooutnow)*40)+((atx-rot)+1)] = useselectedtile -- bottom to left
						roomdata[roomy][roomx][((aty-rot)*40)+((atx-tooutnow)+1)] = useselectedtile -- left to top
					end
					
					toout = toout + 1
				end
				
				autocorrectroom()
			elseif selectedtool == 3 then
				cons("Tile clicked spike: " .. atx .. " " .. aty .. ", set to " .. selectedtile .. ", subtool " .. selectedsubtool[3])
				
				if love.mouse.isDown("r") then
					useselectedtile = 0
				else
					if levelmetadata[(roomy)*20 + (roomx+1)].directmode == 0 then
						-- Auto mode
						useselectedtile = 8
					else
						useselectedtile = selectedtile
					end
				end
				
				if selectedsubtool[3] == 1 then
					-- 1 spike
					roomdata[roomy][roomx][(aty*40)+(atx+1)] = useselectedtile
				elseif selectedsubtool[3] == 2 then
					-- <-->
					if issolid(adjtile((aty*40)+(atx+1), 0, 1), ts) then
						-- There's a solid block below this spike.
						spikes_floor_left(atx, aty, tilesetblocks[selectedtileset].tileimg)
						spikes_floor_right(atx, aty, tilesetblocks[selectedtileset].tileimg)
					elseif issolid(adjtile((aty*40)+(atx+1), 0, -1), ts) then
						-- There's a solid block above this spike.
						spikes_ceiling_right(atx, aty, tilesetblocks[selectedtileset].tileimg)
						spikes_ceiling_left(atx, aty, tilesetblocks[selectedtileset].tileimg)
					elseif issolid(adjtile((aty*40)+(atx+1), -1, 0), ts) then
						-- There's a solid block to the left of this spike.
						spikes_leftwall_up(atx, aty, tilesetblocks[selectedtileset].tileimg)
						spikes_leftwall_down(atx, aty, tilesetblocks[selectedtileset].tileimg)
					elseif issolid(adjtile((aty*40)+(atx+1), 1, 0), ts) then
						-- There's a solid block to the left of this spike.
						spikes_rightwall_down(atx, aty, tilesetblocks[selectedtileset].tileimg)
						spikes_rightwall_up(atx, aty, tilesetblocks[selectedtileset].tileimg)
					else
						-- No solid blocks directly surrounding this spike
						
					end
				elseif selectedsubtool[3] == 3 then
					-- <--
					if issolid(adjtile((aty*40)+(atx+1), 0, 1), ts) then
						-- There's a solid block below this spike.
						spikes_floor_left(atx, aty, tilesetblocks[selectedtileset].tileimg)
					elseif issolid(adjtile((aty*40)+(atx+1), 0, -1), ts) then
						-- There's a solid block above this spike.
						spikes_ceiling_right(atx, aty, tilesetblocks[selectedtileset].tileimg)
					elseif issolid(adjtile((aty*40)+(atx+1), -1, 0), ts) then
						-- There's a solid block to the left of this spike.
						spikes_leftwall_up(atx, aty, tilesetblocks[selectedtileset].tileimg)
					elseif issolid(adjtile((aty*40)+(atx+1), 1, 0), ts) then
						-- There's a solid block to the left of this spike.
						spikes_rightwall_down(atx, aty, tilesetblocks[selectedtileset].tileimg)
					else
						-- No solid blocks directly surrounding this spike
						
					end
				elseif selectedsubtool[3] == 4 then
					-- -->
					if issolid(adjtile((aty*40)+(atx+1), 0, 1), ts) then
						-- There's a solid block below this spike.
						spikes_floor_right(atx, aty, tilesetblocks[selectedtileset].tileimg)
					elseif issolid(adjtile((aty*40)+(atx+1), 0, -1), ts) then
						-- There's a solid block above this spike.
						spikes_ceiling_left(atx, aty, tilesetblocks[selectedtileset].tileimg)
					elseif issolid(adjtile((aty*40)+(atx+1), -1, 0), ts) then
						-- There's a solid block to the left of this spike.
						spikes_leftwall_down(atx, aty, tilesetblocks[selectedtileset].tileimg)
					elseif issolid(adjtile((aty*40)+(atx+1), 1, 0), ts) then
						-- There's a solid block to the left of this spike.
						spikes_rightwall_up(atx, aty, tilesetblocks[selectedtileset].tileimg)
					else
						-- No solid blocks directly surrounding this spike
						
					end
				end
				
				autocorrectroom()
			elseif love.mouse.isDown("l") and not mousepressed and selectedtool == 4 then
				-- Trinket
				if (count.trinkets == 20) and s.allowlimitbypass then
					dialog.new(L.MAXTRINKETS_BYPASS, "", 1, 3, 4)
				elseif count.trinkets >= 20  and not s.allowlimitbypass then
					dialog.new(L.MAXTRINKETS, "", 1, 1, 0)
				else
					-- Same code in dialog.lua for accepting adding a 21st trinket
					cons("Trinket: " .. atx .. " " .. aty)
					
					table.insert(entitydata,
						{
						x = 40*roomx + atx,
						y = 30*roomy + aty,
						t = 9,
						p1 = 0, p2 = 0, p3 = 0, p4 = 0, p5 = 320, p6 = 240,
						data = ""
						})
						
					count.trinkets = count.trinkets + 1
					count.entities = count.entities + 1
						
					if not love.keyboard.isDown("v") then
						mousepressed = true
					end
				end
			elseif love.mouse.isDown("l") and not mousepressed and selectedtool == 5 then
				-- Checkpoint
				cons("Checkpoint: " .. atx .. " " .. aty)
				
				if selectedsubtool[5] == 1 then
					-- Upright
					table.insert(entitydata,
						{
						x = 40*roomx + atx,
						y = 30*roomy + aty,
						t = 10,
						p1 = 1, p2 = 0, p3 = 0, p4 = 0, p5 = 320, p6 = 240,
						data = ""
						})
				elseif selectedsubtool[5] == 2 then
					-- Upside down
					table.insert(entitydata,
						{
						x = 40*roomx + atx,
						y = 30*roomy + aty,
						t = 10,
						p1 = 0, p2 = 0, p3 = 0, p4 = 0, p5 = 320, p6 = 240,
						data = ""
						})
				end
				
				count.entities = count.entities + 1
				
				mousepressed = true
			elseif love.mouse.isDown("l") and not mousepressed and selectedtool == 6 then
				-- Disappearing platform
				cons("Disappear: " .. atx .. " " .. aty)
				
				table.insert(entitydata,
					{
					x = 40*roomx + atx,
					y = 30*roomy + aty,
					t = 3,
					p1 = 0, p2 = 0, p3 = 0, p4 = 0, p5 = 320, p6 = 240,
					data = ""
					})
					
				count.entities = count.entities + 1
					
				mousepressed = true
			elseif love.mouse.isDown("l") and not mousepressed and selectedtool == 7 then
				-- Conveyor
				cons("Conveyor: " .. atx .. " " .. aty)
				
				table.insert(entitydata,
					{
					x = 40*roomx + atx,
					y = 30*roomy + aty,
					t = 2,
					p1 = 4+selectedsubtool[7], p2 = 0, p3 = 0, p4 = 0, p5 = 320, p6 = 240,
					data = ""
					})
				
				count.entities = count.entities + 1
				
				mousepressed = true
			elseif love.mouse.isDown("l") and not mousepressed and selectedtool == 8 then
				-- Moving platform
				cons("Moving: " .. atx .. " " .. aty)
				
				table.insert(entitydata,
					{
					x = 40*roomx + atx,
					y = 30*roomy + aty,
					t = 2,
					p1 = -1+selectedsubtool[8], p2 = 0, p3 = 0, p4 = 0, p5 = 320, p6 = 240,
					data = ""
					})
				
				count.entities = count.entities + 1
				
				mousepressed = true
			elseif love.mouse.isDown("l") and not mousepressed and selectedtool == 9 then
				-- Enemy
				cons("Enemy: " .. atx .. " " .. aty)
				
				table.insert(entitydata,
					{
					x = 40*roomx + atx,
					y = 30*roomy + aty,
					t = 1,
					p1 = -1+selectedsubtool[9], p2 = 0, p3 = 0, p4 = 0, p5 = 320, p6 = 240,
					data = ""
					})
				
				count.entities = count.entities + 1
				
				mousepressed = true
			elseif love.mouse.isDown("l") and not mousepressed and selectedtool == 10 then
				-- Gravity line
				if not issolid(roomdata[roomy][roomx][(aty*40)+(atx+1)], usedtilesets[levelmetadata[(roomy)*20 + (roomx+1)].tileset], true) then
					cons("Gravity line: " .. atx .. " " .. aty)
					
					if selectedsubtool[10] == 1 then
						-- Horizontal
						
						-- Backtrack to see what tile is solid
						for bt = atx, 0, -1 do
							if issolid(roomdata[roomy][roomx][(aty*40)+(bt)], usedtilesets[levelmetadata[(roomy)*20 + (roomx+1)].tileset], true) then
								startat = bt
								break
							end
						end
						if startat == nil then
							startat = -1
						end
						
						-- Now to see how long it should be!
						for ft = startat+1, 40 do
							if issolid(roomdata[roomy][roomx][(aty*40)+(ft)], usedtilesets[levelmetadata[(roomy)*20 + (roomx+1)].tileset], true) then
								linelength = 8 * (ft-startat) - 8
								break
							end
						end
						if linelength == nil then
							linelength = 8 * (42-startat) - 8
						end
						
						table.insert(entitydata,
							{
							x = 40*roomx + atx,
							y = 30*roomy + aty,
							t = 11,
							p1 = 0, p2 = startat, p3 = linelength, p4 = 0, p5 = 320, p6 = 240,
							data = ""
							})
					else
						-- Vertical
						
						-- Backtrack to see what tile is solid
						for bt = aty, 0, -1 do
							--cons("Checking " .. (bt*40)+(atx+1) .. " " .. bt .. " " .. atx)
							if issolid(roomdata[roomy][roomx][(bt*40)+(atx+1)], usedtilesets[levelmetadata[(roomy)*20 + (roomx+1)].tileset], true) then
								startat = bt+1
								break
							end
						end
						if startat == nil then
							startat = -1
						end
						
						-- Now to see how long it should be!
						for ft = startat+1, 29 do
							--cons("Checking2 " .. (ft*40)+(atx+1) .. " " .. ft .. " " .. atx)
							if issolid(roomdata[roomy][roomx][(ft*40)+(atx+1)], usedtilesets[levelmetadata[(roomy)*20 + (roomx+1)].tileset], true) then
								linelength = 8 * (ft-startat)
								break
							end
						end
						if linelength == nil then
							linelength = 8 * (32-startat) - 8
						end
						
						table.insert(entitydata,
							{
							x = 40*roomx + atx,
							y = 30*roomy + aty,
							t = 11,
							p1 = 1, p2 = startat, p3 = linelength, p4 = 0, p5 = 320, p6 = 240,
							data = ""
							})
					end
					
					startat = nil
					linelength = nil
				end
				
				count.entities = count.entities + 1
				
				mousepressed = true
			elseif love.mouse.isDown("l") and not mousepressed and selectedtool == 11 then
				-- Roomtext
				if editingroomtext > 0 then
					-- We were already typing a text!
					endeditingroomtext()
				end
				
				cons("Roomtext: " .. atx .. " " .. aty)
				
				table.insert(entitydata,
					{
					x = 40*roomx + atx,
					y = 30*roomy + aty,
					t = 17,
					p1 = 0, p2 = 0, p3 = 0, p4 = 0, p5 = 320, p6 = 240,
					data = L.BUG
					})
					
				editingroomtext = #entitydata
				makescriptroomtext = false
				startinput()
				
				count.entities = count.entities + 1
				
				mousepressed = true
			elseif love.mouse.isDown("l") and not mousepressed and selectedtool == 12 then
				-- Terminal
				if editingroomtext > 0 then
					-- We were already typing a text!
					endeditingroomtext()
				end
				
				cons("Terminal: " .. atx .. " " .. aty)
				
				table.insert(entitydata,
					{
					x = 40*roomx + atx,
					y = 30*roomy + aty,
					t = 18,
					p1 = 0, p2 = 0, p3 = 0, p4 = 0, p5 = 320, p6 = 240,
					data = "stop"
					})
					
				editingroomtext = #entitydata
				makescriptroomtext = true
				startinput()
				
				count.entities = count.entities + 1
				
				mousepressed = true
			elseif love.mouse.isDown("l") and not mousepressed and selectedtool == 13 then
				-- Script box
				cons("Script box: " .. atx .. " " .. aty)
				
				--dialog.new("Dav needs to work on this! Would be nice if you could use both VVVVV's way of clicking the top left corner and then the bottom right corner, or you can simply drag a box.", "", 1, 1, 0)
				
				-- Subtool is changed in the background
				if selectedsubtool[13] == 1 then
					-- Placing top left corner.
					table.insert(entitydata,
						{
						x = 40*roomx + atx,
						y = 30*roomy + aty,
						t = 19,
						p1 = 0, p2 = 0, p3 = 0, p4 = 0, p5 = 320, p6 = 240,
						data = "stop"
						})
					editingsboxid = #entitydata
					selectedsubtool[13] = 2
					
					mousepressed = true
				elseif selectedsubtool[13] == 2 and (entitydata[editingsboxid] ~= nil) then
					-- Placing bottom right corner
					entitydata[editingsboxid].p1 = math.max((40*roomx + atx) - entitydata[editingsboxid].x + 1, 1)
					entitydata[editingsboxid].p2 = math.max((30*roomy + aty) - entitydata[editingsboxid].y + 1, 1)
					editingsboxid = nil
					selectedsubtool[13] = 1
					
					editingroomtext = #entitydata
					makescriptroomtext = true
					startinput()
					
					mousepressed = true
				else
					--dialog.new("What did you do?!\n\n(script box out of range subtool)", "", 1, 1, 0)
					
					cons("You tried placing the bottom right of a non-existant script box!")
					
					editingsboxid = nil
					selectedsubtool[13] = 1
				end
				
				count.entities = count.entities + 1
				
				--mousepressed = true
			elseif love.mouse.isDown("l") and not mousepressed and selectedtool == 14 then
				-- Warp token
				cons("Warp token: " .. atx .. " " .. aty)
				
				--dialog.new("Dav needs to work on this!", "", 1, 1, 0)
				
				if selectedsubtool[14] == 1 or (selectedsubtool[14] == 2 and entitydata[warpid] == nil) then
					-- Placing entrance.
					table.insert(entitydata,
						{
						x = 40*roomx + atx,
						y = 30*roomy + aty,
						t = 13,
						p1 = 0, p2 = 0, p3 = 0, p4 = 0, p5 = 320, p6 = 240,
						data = ""
						})
					warpid = #entitydata
					selectedsubtool[14] = 2
				elseif selectedsubtool[14] == 2 then
					-- Placing exit
					entitydata[warpid].p1 = 40*roomx + atx
					entitydata[warpid].p2 = 30*roomy + aty
					warpid = nil
					selectedsubtool[14] = 1
				else
					dialog.new("What did you do?!\n\n(warp token out of range subtool)", "", 1, 1, 0)
				end
				
				count.entities = count.entities + 1
				
				mousepressed = true
			elseif love.mouse.isDown("l") and not mousepressed and selectedtool == 15 then
				-- Warp line
				cons("Warp line: " .. atx .. " " .. aty)

				if not issolid(roomdata[roomy][roomx][(aty*40)+(atx+1)], usedtilesets[levelmetadata[(roomy)*20 + (roomx+1)].tileset], true) then
					cons("Warp line: " .. atx .. " " .. aty)
					
					if atx == 0 or atx == 39 then
						-- Vertical left or right, type 0 or 1
						
						-- Backtrack to see what tile is solid
						for bt = aty, 0, -1 do
							--cons("Checking " .. (bt*40)+(atx+1) .. " " .. bt .. " " .. atx)
							if issolid(roomdata[roomy][roomx][(bt*40)+(atx+1)], usedtilesets[levelmetadata[(roomy)*20 + (roomx+1)].tileset], true) then
								startat = bt+1
								break
							end
						end
						if startat == nil then
							startat = -1
						end
						
						-- Now to see how long it should be!
						for ft = startat+1, 29 do
							--cons("Checking2 " .. (ft*40)+(atx+1) .. " " .. ft .. " " .. atx)
							if issolid(roomdata[roomy][roomx][(ft*40)+(atx+1)], usedtilesets[levelmetadata[(roomy)*20 + (roomx+1)].tileset], true) then
								linelength = 8 * (ft-startat)
								break
							end
						end
						if linelength == nil then
							linelength = 8 * (32-startat) - 8
						end
						
						table.insert(entitydata,
							{
							x = 40*roomx + atx,
							y = 30*roomy + aty,
							t = 50,
							p1 = (atx == 0 and 0 or 1), p2 = startat, p3 = linelength, p4 = 0, p5 = 320, p6 = 240,
							data = ""
							})
						count.entities = count.entities + 1
						
						mousepressed = true
					elseif aty == 0 or aty == 29 then
						-- Horizontal top or bottom, type 2 or 3
						
						-- Backtrack to see what tile is solid
						for bt = atx, 0, -1 do
							if issolid(roomdata[roomy][roomx][(aty*40)+(bt)], usedtilesets[levelmetadata[(roomy)*20 + (roomx+1)].tileset], true) then
								startat = bt
								break
							end
						end
						if startat == nil then
							startat = -1
						end
						
						-- Now to see how long it should be!
						for ft = startat+1, 40 do
							if issolid(roomdata[roomy][roomx][(aty*40)+(ft)], usedtilesets[levelmetadata[(roomy)*20 + (roomx+1)].tileset], true) then
								linelength = 8 * (ft-startat) - 8
								break
							end
						end
						if linelength == nil then
							linelength = 8 * (42-startat) - 8
						end
						
						table.insert(entitydata,
							{
							x = 40*roomx + atx,
							y = 30*roomy + aty,
							t = 50,
							p1 = (aty == 0 and 2 or 3), p2 = startat, p3 = linelength, p4 = 0, p5 = 320, p6 = 240,
							data = ""
							})
						count.entities = count.entities + 1
						
						mousepressed = true
					end
					
					startat = nil
					linelength = nil
				end

			elseif love.mouse.isDown("l") and not mousepressed and selectedtool == 16 then
				-- Rescuable crewmate				
				--dialog.new("Crewmates don't feel like being lost yet.", "", 1, 1, 0)
				
				if (count.crewmates == 20) and s.allowlimitbypass then
					dialog.new(L.MAXCREWMATES_BYPASS, "NO QUESTION ID YET!", 1, 3, 0)
				elseif count.crewmates >= 20  and not s.allowlimitbypass then
					dialog.new(L.MAXCREWMATES, "", 1, 1, 0)
				else
					-- Same code in dialog.lua for accepting adding a 21st trinket
					cons("Rescuable crewmate: " .. atx .. " " .. aty .. ", " .. selectedsubtool[selectedtool] .. " " .. anythingbutnil(({1, 2, 3, 4, 5, 0})[selectedsubtool[selectedtool]]))
				
					table.insert(entitydata,
						{
						x = 40*roomx + atx,
						y = 30*roomy + aty,
						t = 15,
						p1 = ({1, 2, 3, 4, 5, 0})[selectedsubtool[selectedtool]], p2 = 0, p3 = 0, p4 = 0, p5 = 320, p6 = 240,
						data = ""
						})
						
					count.crewmates = count.crewmates + 1
					count.entities = count.entities + 1
				end
				
				mousepressed = true
			elseif love.mouse.isDown("l") and not mousepressed and selectedtool == 17 then
				-- Start point
				cons("Start point: " .. atx .. " " .. aty)
				
				-- First remove the old one, but check first
				if (count.startpoint ~= nil) and ( (entitydata[count.startpoint] == nil) or (entitydata[count.startpoint].t ~= 16) ) then
					dialog.new(L.STARTPOINTNOLONGERFOUND .. "\n\nEntity #" .. count.startpoint .. " is '" .. anythingbutnil(anythingbutnil(entitydata[count.startpoint]).t) .. "'.", "", 1, 1, 0)
				else
					if count.startpoint ~= nil then
						cons("Old start point at " .. entitydata[count.startpoint].x .. " " .. entitydata[count.startpoint].y .. " removed")
						entitydata[count.startpoint] = nil
					else
						count.entities = count.entities + 1
					end
					
					table.insert(entitydata,
						{
						x = 40*roomx + atx,
						y = 30*roomy + aty,
						t = 16,
						p1 = selectedsubtool[17]-1, p2 = 0, p3 = 0, p4 = 0, p5 = 320, p6 = 240,
						data = ""
						})
						
					count.startpoint = #entitydata
				end
				
				mousepressed = true
			elseif love.mouse.isDown("l") and not mousepressed then
				dialog.new(L.UNSUPPORTEDTOOL .. anythingbutnil(selectedtool), "", 1, 1, 0)
			end
		--[[
		elseif nodialog and love.mouse.isDown("r") and mouseon(64+64, 0, 639, 480) then
			atx = math.floor((love.mouse.getX()-64-64) / 16)
			aty = math.floor((love.mouse.getY()) / 16)
			
			cons("Tile cleared: " .. atx .. " " .. aty)
			
			roomdata[roomy][roomx][(aty*40)+(atx+1)] = 0
		]]
		elseif nodialog and love.mouse.isDown("m") and mouseon(64+64, 0, 639, 480) and tilespicker and not tilescreator and levelmetadata[(roomy)*20 + (roomx+1)].directmode == 1 then
			local atx = math.floor((love.mouse.getX()-64-64) / 16)
			local aty = math.floor((love.mouse.getY()) / 16)
		
			cons("Tile selected: " .. (aty*40)+(atx+1)-1)
		
			selectedtile = (aty*40)+(atx+1)-1
		elseif nodialog and love.mouse.isDown("m") and mouseon(64+64, 0, 639, 480) and selectedtool <= 3 then --and levelmetadata[(roomy)*20 + (roomx+1)].directmode == 1
			local atx = math.floor((love.mouse.getX()-64-64) / 16)
			local aty = math.floor((love.mouse.getY()) / 16)
			
			selectedtile = roomdata[roomy][roomx][(aty*40)+(atx+1)]
		end
		
		if tilespicker then
			displaytilespicker(64+64, 0, tilesetnames[usedtilesets[levelmetadata[(roomy)*20 + (roomx+1)].tileset]])
		else
			love.graphics.draw(bggrid, 64+64, 0)
			--[[
			cons("Using room " .. (roomy)*20 + (roomx+1))
			for k,v in pairs(levelmetadata[(roomy)*20 + (roomx+1)]) do
				cons(k .. "->" .. v)
			end
			]]
			if love.keyboard.isDown("n") and editingroomtext == 0 then
				love.graphics.setFont(tinynumbers)
			end
			-- Display the room now including its entities, but make sure gravity lines and such don't go offscreen
			love.graphics.setScissor(64+64, 0, 640, 480)
			displayroom(64+64, 0, roomdata[roomy][roomx], levelmetadata[(roomy)*20 + (roomx+1)])
			
			-- Display indicators for tiles in adjacent rooms
			if s.adjacentroomlines then
				local roomupW, roomleftW, roomrightW, roomdownW = false, false, false, false
				local roomup, roomleft, roomright, roomdown
			
				-- Room up. I now notice the bug in the code for switching rooms but it works 100% which is kind of unique to have
				if levelmetadata[roomy*20 + (roomx+1)].warpdir == 2 or levelmetadata[roomy*20 + (roomx+1)].warpdir == 3 then
					-- Use this room because it warps.
					roomup = roomy
					roomupW = true
				else
					if roomy+1 <= 1 then
						roomup = metadata.mapheight-1
					else
						roomup = roomy - 1
					end
				end
				-- Room left
				if levelmetadata[roomy*20 + (roomx+1)].warpdir == 1 or levelmetadata[roomy*20 + (roomx+1)].warpdir == 3 then
					-- Use this room because it warps.
					roomleft = roomx
					roomleftW = true
				else
					if roomx+1 <= 1 then
						roomleft = metadata.mapwidth-1
					else
						roomleft = roomx - 1
					end
				end
				-- Room right
				if levelmetadata[roomy*20 + (roomx+1)].warpdir == 1 or levelmetadata[roomy*20 + (roomx+1)].warpdir == 3 then
					-- Use this room because it warps.
					roomright = roomx
					roomrightW = true
				else
					if roomx+1 >= metadata.mapwidth then
						roomright = 0
					else
						roomright = roomx + 1
					end
				end
				-- Room down
				if levelmetadata[roomy*20 + (roomx+1)].warpdir == 2 or levelmetadata[roomy*20 + (roomx+1)].warpdir == 3 then
					-- Use this room because it warps.
					roomdown = roomy
					roomdownW = true
				else
					if roomy+1 >= metadata.mapheight then
						roomdown = 0
					else
						roomdown = roomy + 1
					end
				end
				
				-- Up
				for t = 0, 39 do
					if issolid(roomdata[roomup][roomx][t+1161], usedtilesets[levelmetadata[(roomup)*20 + (roomx+1)].tileset]) then
						--love.graphics.line(64+64+(t*16) +1, 0, 64+64+(t*16)+15, 0)
						--love.graphics.line(64+64+(t*16) +1, 1, 64+64+(t*16)+15, 1)
						
						if roomupW then
							love.graphics.setColor(0, 192, 255)
						end
						
						love.graphics.draw(sideimg, sideline[1], 64+64+(t*16), 0, 0, 2)
						
						if roomupW then
							love.graphics.setColor(255, 255, 255)
						end
					elseif not roomupW and ( (levelmetadata[(roomup)*20 + (roomx+1)].warpdir == 2) or (levelmetadata[(roomup)*20 + (roomx+1)].warpdir == 3) ) then
						love.graphics.draw(smallsideimg, smallsideline[1], 64+64+(t*16), 0, 0, 1)
					end
				end
				-- Left
				for t = 1, 30 do
					if issolid(roomdata[roomy][roomleft][t*40], usedtilesets[levelmetadata[(roomy)*20 + (roomleft+1)].tileset]) then
						--love.graphics.line(64+64, (t-1)*16 +1, 64+64, (t-1)*16 +15)
						--love.graphics.line(64+64+1, (t-1)*16 +1, 64+64+1, (t-1)*16 +15)
						
						if roomleftW then
							love.graphics.setColor(0, 192, 255)
						end
						
						love.graphics.draw(sideimg, sideline[2], 64+64, (t-1)*16, 0, 2)
						
						if roomleftW then
							love.graphics.setColor(255, 255, 255)
						end
					elseif not roomleftW and ( (levelmetadata[(roomy)*20 + (roomleft+1)].warpdir == 1) or (levelmetadata[(roomy)*20 + (roomleft+1)].warpdir == 3) ) then
						love.graphics.draw(smallsideimg, smallsideline[2], 64+64, (t-1)*16, 0, 1)
					end
				end
				-- Right
				for t = 0, 29 do
					if issolid(roomdata[roomy][roomright][t*40 + 1], usedtilesets[levelmetadata[(roomy)*20 + (roomright+1)].tileset]) then
					
						if roomrightW then
							love.graphics.setColor(0, 192, 255)
						end
					
						love.graphics.draw(sideimg, sideline[3], 64+64+(39*16), t*16, 0, 2)
						
						if roomrightW then
							love.graphics.setColor(255, 255, 255)
						end
						
					elseif not roomrightW and ( (levelmetadata[(roomy)*20 + (roomright+1)].warpdir == 1) or (levelmetadata[(roomy)*20 + (roomright+1)].warpdir == 3) ) then
						love.graphics.draw(smallsideimg, smallsideline[3], 64+64+(39*16), t*16, 0, 1)
					end
				end
				-- Down
				for t = 0, 39 do
					if issolid(roomdata[roomdown][roomx][t + 1], usedtilesets[levelmetadata[(roomdown)*20 + (roomx+1)].tileset]) then
					
						if roomdownW then
							love.graphics.setColor(0, 192, 255)
						end
					
						love.graphics.draw(sideimg, sideline[4], 64+64+(t*16), 29*16, 0, 2)
						
						if roomdownW then
							love.graphics.setColor(255, 255, 255)
						end
						
					elseif not roomdownW and ( (levelmetadata[(roomdown)*20 + (roomx+1)].warpdir == 2) or (levelmetadata[(roomdown)*20 + (roomx+1)].warpdir == 3) ) then
						love.graphics.draw(smallsideimg, smallsideline[4], 64+64+(t*16), 29*16, 0, 1)
					end
				end
			end
			
			love.graphics.setFont(font8)
			displayentities(64+64, 0, roomx, roomy)
			love.graphics.setScissor()
			
			-- Does this room have a name, perhaps?
			if temporaryroomnametimer > 0 then
				-- Oh wait, we're displaying a message as room name!
				love.graphics.setColor(160,160,0,128)
				love.graphics.rectangle("fill", 64+64, 29*16, 40*16, 16)
				love.graphics.setColor(255,255,255,255)
				love.graphics.setFont(font16)
				love.graphics.printf(temporaryroomname, 64+64, 29*16 +3, 40*16, "center")
				love.graphics.setFont(font8)
			elseif editingroomname then
				-- We're editing this room name!
				love.graphics.setColor(128,128,128,128)
				love.graphics.rectangle("fill", 64+64, 29*16, 40*16, 16)
				love.graphics.setColor(255,255,255,255)
				love.graphics.setFont(font16)
				love.graphics.printf(input .. __, 64+64, 29*16 +3, 40*16, "center")
				love.graphics.setFont(font8)
			elseif levelmetadata[(roomy)*20 + (roomx+1)].roomname ~= "" then
				-- Display it
				love.graphics.setColor(0,0,0,128)
				love.graphics.rectangle("fill", 64+64, 29*16, 40*16, 16)
				love.graphics.setColor(255,255,255,255)
				love.graphics.setFont(font16)
				love.graphics.printf(levelmetadata[(roomy)*20 + (roomx+1)].roomname, 64+64, 29*16 +3, 40*16, "center")
				love.graphics.setFont(font8)
			end
		end
		
		-- Now display the cursor. If it's on the level
		if nodialog and mouseon(64+64, 0, 639, 480) then
			cursorx = math.floor((love.mouse.getX()-64-64) / 16)
			cursory = math.floor((love.mouse.getY()) / 16)
			
			-- Are we supposed to display a special cursor shape?
			if tilespicker then
				if levelmetadata[(roomy)*20 + (roomx+1)].directmode == 0 then else
					-- Just one tile, but only in manual/direct mode.
					love.graphics.draw(cursorimg[0], (cursorx*16)+64+64, (cursory*16))
				end
			elseif selectedtool == 1 or selectedtool == 2 then
				-- Wall and background have different kinds of possible cursor shapes
				if selectedsubtool[selectedtool] == 1 then
					-- Just a regular cursor
					displayalphatile(0, 0, 0, 0)
					love.graphics.draw(cursorimg[0], (cursorx*16)+64+64, (cursory*16))
				elseif selectedsubtool[selectedtool] == 2 then
					-- 3x3
					displayalphatile(1, 1, 2, 2)
					displayshapedcursor(1, 1, 1, 1)
				elseif selectedsubtool[selectedtool] == 3 then
					-- 5x5
					displayalphatile(2, 2, 4, 4)
					displayshapedcursor(2, 2, 2, 2)
				elseif selectedsubtool[selectedtool] == 4 then
					-- 7x7
					displayalphatile(3, 3, 6, 6)
					displayshapedcursor(3, 3, 3, 3)
				elseif selectedsubtool[selectedtool] == 5 then
					-- 9x9
					displayalphatile(4, 4, 8, 8)
					displayshapedcursor(4, 4, 4, 4)
				elseif selectedsubtool[selectedtool] == 6 then
					-- horizontal fill
					displayalphatile_hor()
					love.graphics.draw(cursorimg[1], 64+64, (cursory*16))
					love.graphics.draw(cursorimg[2], 64+64+(39*16), (cursory*16))
					love.graphics.draw(cursorimg[3], 64+64, (cursory*16))
					love.graphics.draw(cursorimg[4], 64+64+(39*16), (cursory*16))
				elseif selectedsubtool[selectedtool] == 7 then
					-- vertical fill
					displayalphatile_ver()
					love.graphics.draw(cursorimg[1], 64+64+(cursorx*16), 0)
					love.graphics.draw(cursorimg[2], 64+64+(cursorx*16), 0)
					love.graphics.draw(cursorimg[3], 64+64+(cursorx*16), (29*16))
					love.graphics.draw(cursorimg[4], 64+64+(cursorx*16), (29*16))
				elseif selectedsubtool[selectedtool] == 8 then
					-- Room fill
					displayalphatile_all()
					love.graphics.draw(cursorimg[1], 64+64, 0)
					love.graphics.draw(cursorimg[2], 64+64+(39*16), 0)
					love.graphics.draw(cursorimg[3], 64+64, (29*16))
					love.graphics.draw(cursorimg[4], 64+64+(39*16), (29*16))
				elseif selectedsubtool[selectedtool] == 9 then
					displayalphatile(-1, 0, 0, 0)
					displayalphatile(1, 0, 0, 0)
					displayalphatile(0, -1, 0, 0)
					displayalphatile(0, 1, 0, 0)
					love.graphics.draw(cursorimg[0], (cursorx*16)+64+64-16, (cursory*16))
					love.graphics.draw(cursorimg[0], (cursorx*16)+64+64+16, (cursory*16))
					love.graphics.draw(cursorimg[0], (cursorx*16)+64+64, (cursory*16)-16)
					love.graphics.draw(cursorimg[0], (cursorx*16)+64+64, (cursory*16)+16)
				end
			elseif selectedtool == 3 then
				-- Spike
				displayalphatile(0, 0, 0, 0)
				love.graphics.draw(cursorimg[0], (cursorx*16)+64+64, (cursory*16))
			elseif selectedtool == 4 then
				-- Trinket
				displayshapedcursor(0, 0, 1, 1)
			elseif selectedtool == 5 then
				-- Checkpoint
				displayshapedcursor(0, 0, 1, 1)
			elseif selectedtool == 6 then
				-- Disappearing platform
				displayshapedcursor(0, 0, 3, 0)
			elseif selectedtool == 7 then
				-- Conveyor
				if selectedsubtool[7] <= 2 then
					displayshapedcursor(0, 0, 3, 0)
				else
					displayshapedcursor(0, 0, 7, 0)
				end
			elseif selectedtool == 8 then
				-- Moving platform
				displayshapedcursor(0, 0, 3, 0)
			elseif selectedtool == 9 then
				-- Enemy
				displayshapedcursor(0, 0, 1, 1)
			elseif selectedtool == 12 then
				-- Terminal
				displayshapedcursor(0, 0, 1, 2)
			elseif selectedtool == 14 then
				-- Warp token
				displayshapedcursor(0, 0, 1, 1)
			elseif selectedtool == 16 then
				-- Crewmate
				displayshapedcursor(0, 0, 1, 2)
			elseif selectedtool == 17 then
				-- Start point
				displayshapedcursor(0, 0, 1, 2)
			elseif not (selectedtool == 13 and selectedsubtool[13] == 2) then
				love.graphics.draw(cursorimg[0], (cursorx*16)+64+64, (cursory*16))
			end
		else
			cursorx = "--"
			cursory = "--"
		end
		
		-- We also want the tools on the left. But it's a scrollable area.
		love.graphics.setScissor(16, 16, 32+4, love.graphics.getHeight()-32)
		
		for t = 1, 17 do
			-- love.graphics.rectangle("fill", 16, (16+(48*(t-1)))+lefttoolscroll, 32, 32)
			-- Are we hovering over it? Or maybe even clicking it?
			if not nodialog or ((mouseon(16, 0, 32, 16)) or (mouseon(16, love.graphics.getHeight()-16, 32, 16)) or (not mouseon(16, (16+(48*(t-1)))+lefttoolscroll, 32, 32))) and selectedtool ~= t then
				love.graphics.setColor(255,255,255,128)
			end
			
			if nodialog and love.mouse.isDown("l") and mouseon(16, (16+(48*(t-1)))+lefttoolscroll, 32, 32) and not mouseon(16, 0, 32, 16) and not mouseon(16, love.graphics.getHeight()-16, 32, 16) then
				selectedtool = t
				love.window.setIcon(toolimgicon[t])
			end
			
			if nodialog and love.mouse.isDown("r") and t == 17 and mouseon(16, (16+(48*(t-1)))+lefttoolscroll, 32, 32) and not mouseon(16, 0, 32, 16) and not mouseon(16, love.graphics.getHeight()-16, 32, 16) then
				-- Find the start point
				gotostartpointroom()
			end
			
			if selectedtool == t then
				love.graphics.draw(selectedtoolborder,  16, (16+(48*(t-1)))+lefttoolscroll)
			else
				love.graphics.draw(unselectedtoolborder,  16, (16+(48*(t-1)))+lefttoolscroll)
			end
			
			coorx = 16+2
			coory = (16+2+(48*(t-1)))+lefttoolscroll
			
			love.graphics.draw(toolimg[t], coorx, coory)
			love.graphics.setColor(255,255,255,255)
			
			-- Put the shortcut next to it.
			love.graphics.setFont(tinynumbers)
			love.graphics.print(toolshortcuts[t], coorx-2+32+1, coory)
			love.graphics.setFont(font8)
			
			if nodialog and ((not mouseon(16, 0, 32, 16)) and not (mouseon(16, love.graphics.getHeight()-16, 32, 16)) and (mouseon(16, (16+(48*(t-1)))+lefttoolscroll, 32, 32))) then
				-- Ugh this code but we're hovering over it. So display a tooltip, but don't let it get snipped away by the scissors.
				love.graphics.setScissor()
				love.graphics.setColor(128,128,128,192)
				love.graphics.rectangle("fill", love.mouse.getX()+15, love.mouse.getY()-10, font8:getWidth(toolnames[t]), 8) -- string.len(toolnames[t])*8
				love.graphics.setColor(255,255,255,255)
				love.graphics.print(toolnames[t], love.mouse.getX()+16, love.mouse.getY()-8)
				love.graphics.setScissor(16, 16, 32+4, love.graphics.getHeight()-32)
			end
		end
		
		toolfinish({coorx, coory, selectedtool})
		
		-- We're done with that now, but now we have an area with subtools.
		love.graphics.setScissor(16+64, 16, 32, love.graphics.getHeight()-32)
		
		for k,v in pairs(subtoolimgs[selectedtool]) do
			-- Are we hovering over it? Or maybe even clicking it?
			if not nodialog or (selectedtool == 14 and selectedsubtool[selectedtool] ~= k) then
				love.graphics.setColor(255,255,255,128)
			elseif not nodialog or ((mouseon(16+64, 0, 32, 16)) or (mouseon(16+64, love.graphics.getHeight()-16, 32, 16)) or (not mouseon(16+64, (16+(48*(k-1)))+leftsubtoolscroll, 32, 32))) and selectedsubtool[selectedtool] ~= k then
				love.graphics.setColor(255,255,255,128)
			end
			
			if nodialog and love.mouse.isDown("l") and mouseon(16+64, (16+(48*(k-1)))+leftsubtoolscroll, 32, 32) and not mouseon(16+64, 0, 32, 16) and not mouseon(16+64, love.graphics.getHeight()-16, 32, 16) and selectedtool ~= 14 then
				selectedsubtool[selectedtool] = k
			end
			
			if selectedsubtool[selectedtool] == k then
				love.graphics.draw(selectedtoolborder,  16+64, (16+(48*(k-1)))+leftsubtoolscroll)
			else
				love.graphics.draw(unselectedtoolborder,  16+64, (16+(48*(k-1)))+leftsubtoolscroll)
			end
			
			coorx = 16+64+2
			coory = (16+2+(48*(k-1)))+leftsubtoolscroll
			
			-- v = subtoolimgs[selectedtool][k]
			love.graphics.draw(v, coorx, coory)
			love.graphics.setColor(255,255,255,255)
			
			if nodialog and ((not mouseon(16+64, 0, 32, 16)) and not (mouseon(16+64, love.graphics.getHeight()-16, 32, 16)) and (mouseon(16+64, (16+(48*(k-1)))+leftsubtoolscroll, 32, 32))) then
				-- Ugh this code but we're hovering over it. So display a tooltip, but don't get snipped away by the scissors.
				love.graphics.setScissor()
				love.graphics.setColor(128,128,128,192)
				love.graphics.rectangle("fill", love.mouse.getX()+15, love.mouse.getY()-10, font8:getWidth(anythingbutnil(subtoolnames[selectedtool][k])), 8) -- string.len(anythingbutnil(subtoolnames[selectedtool][k]))*8
				love.graphics.setColor(255,255,255,255)
				love.graphics.print(anythingbutnil(subtoolnames[selectedtool][k]), love.mouse.getX()+16, love.mouse.getY()-8)
				love.graphics.setScissor(16+64, 16, 32, love.graphics.getHeight()-32)
			end
		end
		
		-- We're done with the scrollable menus now.
		love.graphics.setScissor()
		
		-- But we still want to be able to scroll!
		hoverdraw(scrollup, 16, 0, 32, 16)
		hoverdraw(scrolldn, 16, love.graphics.getHeight()-16, 32, 16)
		
		-- Are we clicking them?
		if nodialog and love.mouse.isDown("l") and mouseon(16, 0, 32, 16) then
			lefttoolscroll = lefttoolscroll + 4
			lefttoolscrollbounds()
		elseif nodialog and love.mouse.isDown("l") and mouseon(16, love.graphics.getHeight()-16, 32, 16) then
			lefttoolscroll = lefttoolscroll - 4
			lefttoolscrollbounds()
		end
		
		-- Now stuff on the right.
		hoverdraw(helpbtn, love.graphics.getWidth()-128+8, 8, 16, 16, 1)
		hoverdraw(newbtn, love.graphics.getWidth()-96, 0, 32, 32, 2)
		hoverdraw(loadbtn, love.graphics.getWidth()-64, 0, 32, 32, 2)
		hoverdraw(savebtn, love.graphics.getWidth()-32, 0, 32, 32, 2)
		
		-- Now for the other buttons - about this variable, I can hardcode it again later.
		local buttonspacing = 20 --24
		
		hoverrectangle(128,128,128,128, love.graphics.getWidth()-(128-8), 40, 128-16, 16)
		love.graphics.printf(L.VEDOPTIONS, love.graphics.getWidth()-(128-8), (40)+4+2, 128-16, "center")
		hoverrectangle(128,128,128,128, love.graphics.getWidth()-(128-8), 40+buttonspacing, 128-16, 16)
		love.graphics.printf(L.LEVELOPTIONS, love.graphics.getWidth()-(128-8), (40+buttonspacing)+4+2, 128-16, "center")
		hoverrectangle(128,128,128,128, love.graphics.getWidth()-(128-8), 40+(2*buttonspacing), 128-16, 16)
		love.graphics.printf(L.MAP, love.graphics.getWidth()-(128-8), (40+(2*buttonspacing))+4+2, 128-16, "center")
		hoverrectangle(128,128,128,128, love.graphics.getWidth()-(128-8), 40+(3*buttonspacing), 128-16, 16)
		love.graphics.printf(L.SCRIPTS, love.graphics.getWidth()-(128-8), (40+(3*buttonspacing))+4+2, 128-16, "center")
		hoverrectangle(128,128,128,128, love.graphics.getWidth()-(128-8), 40+(4*buttonspacing), 128-16, 16)
		love.graphics.printf(L.SEARCH, love.graphics.getWidth()-(128-8), (40+(4*buttonspacing))+4+2, 128-16, "center")
		hoverrectangle(128,128,128,128, love.graphics.getWidth()-(128-8), 40+(5*buttonspacing), 128-16, 16)
		love.graphics.printf("Level notepad", love.graphics.getWidth()-(128-8), (40+(5*buttonspacing))+4+2, 128-16, "center") -- Test
		hoverrectangle(128,128,128,128, love.graphics.getWidth()-(128-8), 40+(6*buttonspacing), 128-16, 16)
		love.graphics.printf(L.SENDFEEDBACK, love.graphics.getWidth()-(128-8), (40+(6*buttonspacing))+4+2, 128-16, "center") -- Test

		
		hoverrectangle(128,128,128,128, love.graphics.getWidth()-(128-8), love.graphics.getHeight()-(6*16)-16-24-12-8-(24*1), 128-16, 16)
		love.graphics.printf((roomoptpage2 and L.BACKB or L.MOREB), love.graphics.getWidth()-(128-8), (love.graphics.getHeight()-(6*16)-16-24-12-8-(24*1))+4+2, 128-16, "center")
		if not roomoptpage2 then
			hoverrectangle(128,128,128,128, love.graphics.getWidth()-(128-8), love.graphics.getHeight()-(6*16)-16-24-12-8-(24*2), 128-16, 16)
			love.graphics.printf((roomoptpage2 and "Button?" or (levelmetadata[(roomy)*20 + (roomx+1)].directmode == 1 and L.MANUALMODE or L.AUTOMODE)), love.graphics.getWidth()-(128-8), (love.graphics.getHeight()-(6*16)-16-24-12-8-(24*2))+4+2, 128-16, "center")
		end
		hoverrectangle(128,128,128,128, love.graphics.getWidth()-(128-8), love.graphics.getHeight()-(6*16)-16-24-12-8-(24*3), 128-16, 16)
		love.graphics.printf((roomoptpage2 and langkeys(L.PLATFORMSPEED, {levelmetadata[(roomy)*20 + (roomx+1)].platv}) or langkeys(L.ENEMYTYPE, {levelmetadata[(roomy)*20 + (roomx+1)].enemytype})), love.graphics.getWidth()-(128-8), (love.graphics.getHeight()-(6*16)-16-24-12-8-(24*3))+4+2, 128-16, "center")
		hoverrectangle(128,128,128,128, love.graphics.getWidth()-(128-8), love.graphics.getHeight()-(6*16)-16-24-12-8-(24*4), 128-16, 16)
		love.graphics.printf((roomoptpage2 and L.PLATFORMBOUNDS or langkeys(L.WARPDIR, {warpdirs[levelmetadata[(roomy)*20 + (roomx+1)].warpdir]})), love.graphics.getWidth()-(128-8), (love.graphics.getHeight()-(6*16)-16-24-12-8-(24*4))+4+2, 128-16, "center")
		hoverrectangle(128,128,128,128, love.graphics.getWidth()-(128-8), love.graphics.getHeight()-(6*16)-16-24-12-8-(24*5), 128-16, 16)
		love.graphics.printf((roomoptpage2 and L.ENEMYBOUNDS or L.ROOMNAME), love.graphics.getWidth()-(128-8), (love.graphics.getHeight()-(6*16)-16-24-12-8-(24*5))+4+2, 128-16, "center")

		love.graphics.printf(L.ROOMOPTIONS, love.graphics.getWidth()-(128-8), (love.graphics.getHeight()-(6*16)-16-24-12-8-(24*6))+4+2+4, 128-16, "center")
		
		-- Well make them actually do something!
		if not mousepressed and nodialog and love.mouse.isDown("l") then
			if mouseon(love.graphics.getWidth()-128+8, 8, 16, 16) then
				-- Help
				--dialog.new("A help thing will appear where you can click a tutorial button and some topic-specific buttons\n\nMaybe add something for being able to click on an area/button on the screen to get help about that", "", 1, 1, 0)
				tostate(15)
			elseif mouseon(love.graphics.getWidth()-96, 0, 32, 32) then
				-- New
				--dialog.new("Making a new level isn't supported yet!", "", 1, 1, 0)
				dialog.new(L.SURENEWLEVEL .. "\n\n(dialog will be save/don't save/cancel later)", "", 1, 3, 7)
			elseif mouseon(love.graphics.getWidth()-64, 0, 32, 32) then
				-- Load. But first ask them if they want to save (make this save/don't save/cancel later, yes/no for now)
				dialog.new(L.SURELOADLEVEL .. "\n\n(dialog will be save/don't save/cancel later)", "", 1, 3, 3)
			elseif mouseon(love.graphics.getWidth()-32, 0, 32, 32) then
				-- Save
				--tostate(8)
				startmultiinput({(editingmap ~= "untitled\n" and editingmap or "")})
				dialog.new(L.ENTERNAMESAVE .. "\n", "", 1, 4, 10)
			elseif mouseon(love.graphics.getWidth()-(128-8), 40, 128-16, 16) then
				-- Ved options
				--dialog.new("Ved options coming soon\n\nFor now, you can edit the config file at %appdata%\\LOVE\\ved\\settings.lua (Windows) or Library/Application Support/LOVE/ved/settings.lua (Mac)\n\nThis dialog is so temporary that I won't bother checking the OS for it to display the correct folder name.", "", 1, 1, 0)
				tostate(13)
			elseif mouseon(love.graphics.getWidth()-(128-8), 40+buttonspacing, 128-16, 16) then
				-- Level options
				startmultiinput({metadata.Title, metadata.Creator, metadata.website, metadata.Desc1, metadata.Desc2, metadata.Desc3, metadata.mapwidth, metadata.mapheight, metadata.levmusic})
				dialog.new(L.OPTNAME .. "\n" .. L.OPTBY .. "\n" .. L.OPTWEBSITE .. "\n\n" .. L.OPTDESC .. "\n\n\n\n" .. L.OPTSIZE .. "\n\n" .. L.OPTMUSIC, "Level options", 1, 4, 5)
			elseif mouseon(love.graphics.getWidth()-(128-8), 40+(2*buttonspacing), 128-16, 16) then
				-- Map
				--dialog.new("Map coming soon", "", 1, 1, 0)
				tostate(12)
			elseif mouseon(love.graphics.getWidth()-(128-8), 40+(3*buttonspacing), 128-16, 16) then
				-- Scripts
				tostate(10)
			elseif mouseon(love.graphics.getWidth()-(128-8), 40+(4*buttonspacing), 128-16, 16) then
				-- Search
				--dialog.new("From here you will be able to search through:\n- Scripts and script names\n- Room names\n- Roomtext\n\n...Am I missing anything?", "", 1, 1, 0)
				tostate(11)
			elseif not mousepressed and mouseon(love.graphics.getWidth()-(128-8), 40+(5*buttonspacing), 128-16, 16) then
				-- Level notepad
				if vedmetadata ~= false then
					tostate(15, nil, vedmetadata.notes)
				else
					--dialog.new(L.METADATAENTITYCREATENOW
					vedmetadata = createmde()
					tostate(15, nil, vedmetadata.notes)
				end
			elseif not mousepressed and mouseon(love.graphics.getWidth()-(128-8), 40+(6*buttonspacing), 128-16, 16) then
				-- Test BUT "SEND FEEDBACK" FOR NOW
				--dialog.new("Auto-creation of a save file for VVVVVV coming soon!", "", 1, 1, 0)
				openurl("http://ved.idea.informer.com/")
				
				mousepressed = true
				
			-- Room options now
			elseif not mousepressed and mouseon(love.graphics.getWidth()-(128-8), love.graphics.getHeight()-(6*16)-16-24-12-8-(24*1), 128-16, 16) then
				-- Bounds
				--dialog.new("Bounds", "", 1, 1, 0)
				roomoptpage2 = not roomoptpage2
				mousepressed = true
			elseif not mousepressed and mouseon(love.graphics.getWidth()-(128-8), love.graphics.getHeight()-(6*16)-16-24-12-8-(24*2), 128-16, 16) then
				-- Direct mode on or off
				if not roomoptpage2 then
					-- Direct mode on or off
					--dialog.new("Direct mode on/off", "", 1, 1, 0)
					changedmode()
				end
				
				mousepressed = true
			elseif not mousepressed and mouseon(love.graphics.getWidth()-(128-8), love.graphics.getHeight()-(6*16)-16-24-12-8-(24*3), 128-16, 16) then
				-- Enemy type // Platform speed
				if not roomoptpage2 then
					-- Enemy type
					--dialog.new("Enemy type", "", 1, 1, 0)
					switchenemies()
				else
					-- Platform speed
					--dialog.new("Platform speed", "", 1, 1, 0)
					levelmetadata[(roomy)*20 + (roomx+1)].platv = cycle(levelmetadata[(roomy)*20 + (roomx+1)].platv, 8, 0)
				end
				
				mousepressed = true
			elseif not mousepressed and mouseon(love.graphics.getWidth()-(128-8), love.graphics.getHeight()-(6*16)-16-24-12-8-(24*4), 128-16, 16) then
				-- Warp dir // Platform bounds
				if not roomoptpage2 then
					-- Warp dir
					changewarpdir()
				else
					-- Platform bounds
					changeplatformbounds()
				end
				
				mousepressed = true
			elseif not mousepressed and mouseon(love.graphics.getWidth()-(128-8), love.graphics.getHeight()-(6*16)-16-24-12-8-(24*5), 128-16, 16) then
				-- Roomname // Enemy bounds
				if not roomoptpage2 then
					-- Roomname
					toggleeditroomname()
				else
					-- Enemy bounds
					changeenemybounds()
				end
				
				mousepressed = true
			end
			
			-- Gewoon overal??
			--mousepressed = true
			-- Nee want dan kun je hieronder geen klikken meer opvangen
		end
		
		if nodialog and love.mouse.isDown("r") and mouseon(love.graphics.getWidth()-(128-8), love.graphics.getHeight()-(6*16)-16-24-12-8-(24*3), 128-16, 16) then
			levelmetadata[(roomy)*20 + (roomx+1)].platv = 4
		end
		
		-- And coordinates.
		love.graphics.print("(" .. (roomx+1) .. "," .. (roomy+1) .. ")", love.graphics.getWidth()-56, love.graphics.getHeight()-16-8)
		
		-- But if we're in the tiles picker instead display the tile we're hovering on!
		if tilespicker then
			if (cursorx ~= "--") and (cursory ~= "--") then
				love.graphics.print(L.TILE .. " " .. (cursory*40)+(cursorx+1)-1, love.graphics.getWidth()-72, love.graphics.getHeight()-8-8)
				love.graphics.print((issolid((cursory*40)+(cursorx+1)-1, usedtilesets[levelmetadata[(roomy)*20 + (roomx+1)].tileset]) and L.SOLID or L.NOTSOLID), love.graphics.getWidth()-72, love.graphics.getHeight()-8)
			else
				love.graphics.print(L.TILE .. " ----", love.graphics.getWidth()-72, love.graphics.getHeight()-8-8)
			end
		else
			love.graphics.print("[" .. cursorx .. "," .. cursory .. "]", love.graphics.getWidth()-56, love.graphics.getHeight()-8-8)
			if (cursorx ~= "--") and (cursory ~= "--") then
				love.graphics.print("<" .. (cursorx*8) .. "," .. (cursory*8) .. ">", love.graphics.getWidth()-56-16, love.graphics.getHeight()-8)
			else
				love.graphics.print("<---,--->", love.graphics.getWidth()-56-16, love.graphics.getHeight()-8)
			end
		end
		
		-- Also display a smaller tiles picker for semi-undirect mode
		love.graphics.rectangle("fill", love.graphics.getWidth()-(7*16)-1, love.graphics.getHeight()-(6*16)-16-24-12-1-8, (6*16)+2, (5*16)+2)
		love.graphics.setColor(0,0,0,255)
		love.graphics.rectangle("fill", love.graphics.getWidth()-(7*16), love.graphics.getHeight()-(6*16)-16-24-12-8, (6*16), (5*16))
		love.graphics.setColor(255,255,255,255)
		if selectedtool <= 3 then
			displaysmalltilespicker(love.graphics.getWidth()-(7*16), love.graphics.getHeight()-(6*16)-16-24-12-8, selectedtileset, selectedcolor)
		end
		
		-- And text below it.
		--[[
		hoverrectangle(128,128,128,128, love.graphics.getWidth()-(7*16)-1, love.graphics.getHeight()-16-32-2, (6*16), 8)
		love.graphics.print("<tiles2.png>", love.graphics.getWidth()-(7*16), love.graphics.getHeight()-16-32)
		]]
		
		hoverrectangle(128,128,128,128, love.graphics.getWidth()-(7*16)-1, love.graphics.getHeight()-16-32-2-12-8, (6*16), 8+4)
		love.graphics.print(" " .. (tilesetblocks[selectedtileset].name ~= nil and tilesetblocks[selectedtileset].name or selectedtileset), love.graphics.getWidth()-(7*16), love.graphics.getHeight()-16-32-12+2-8)
		love.graphics.print(" ", love.graphics.getWidth()-(1.5*16), love.graphics.getHeight()-16-32-12+2-8) -- v
		
		hoverrectangle(128,128,128,128, love.graphics.getWidth()-(7*16)-1, love.graphics.getHeight()-16-24-2-8-8, (6*16), 8+4)
		love.graphics.print(" " .. (tilesetblocks[selectedtileset].colors[selectedcolor].name ~= nil and tilesetblocks[selectedtileset].colors[selectedcolor].name or langkeys(L.TSCOLOR, {selectedcolor})), love.graphics.getWidth()-(7*16), love.graphics.getHeight()-16-24-8+2-8)
		love.graphics.print(" ", love.graphics.getWidth()-(1.5*16), love.graphics.getHeight()-16-24-8+2-8) -- v
		
		if love.mouse.isDown("l") and not mousepressed and mouseon(love.graphics.getWidth()-(7*16)-1, love.graphics.getHeight()-16-32-2-12-8, (6*16), 8+4) then
			-- Switch tileset
			switchtileset()
		elseif love.mouse.isDown("l") and not mousepressed and mouseon(love.graphics.getWidth()-(7*16)-1, love.graphics.getHeight()-16-24-2-8-8, (6*16), 8+4) then
			-- Switch tilecol
			switchtilecol()
		end
		
		hoverrectangle(128,128,128,128, love.graphics.getWidth()-(7*16)-1, love.graphics.getHeight()-16-16-2-4-8, (6*16), 8+4)
		love.graphics.print(tilespicker and L.HIDEALL or L.SHOWALL, love.graphics.getWidth()-(7*16), love.graphics.getHeight()-16-16-4+2-8)
		
		-- Some text below the tiles picker-- how many trinkets and crewmates do we have?
		--love.graphics.printf("Trinkets: " .. anythingbutnil(count.trinkets) .. "/20\nCrewmates: " .. anythingbutnil(count.crewmates) .. "/20", 768, love.graphics.getHeight()-(6*16)-16-24-12-16, 128, "right")
		love.graphics.printf(L.ONETRINKETS .. fixdige(anythingbutnil(count.trinkets), 2, "", "!") .. "/20\n" .. L.ONECREWMATES .. fixdige(anythingbutnil(count.crewmates), 2, "", "!") .. "/20\n" .. L.ONEENTITIES .. fixdig(anythingbutnil(count.entities), 5, ""), 768, love.graphics.getHeight()-16-8, 128, "left")
		
		
		-- Dropdown for tileset?
		--[[
		if dropdown == 1 then
			--[ [
			love.graphics.print([ [-1-
			 Space St.
			-2-
			 Outside
			 Lab
			 Warp Zone
			 Ship] ], love.graphics.getWidth()-(7*16)-1, love.graphics.getHeight()-16-32-2-12-56)
			 ] ]
			if not nodialog then
				rightclickmenu.create({"#1", "Space St.", "#2", "Outside", "Lab", "Warp Zone", "Ship"}, "tileset")
			end
		elseif dropdown == 2 then
		
		end
		]]
		
		-- Do we want to see room metadata?
		if allowdebug and love.keyboard.isDown("f11") then
			
		end
	elseif state == 2 then
		love.graphics.print("Syntax highlighting" .. input .. __, 10, 10)
		
		syntaxhl(input, 10, 30, false, true)
		startinputonce()
	elseif state == 3 then
		love.graphics.setColor(255,255,255,255)
		love.graphics.print(L.SCRIPTEDITOR .. " - " .. scriptname, 8, 8)
		--love.graphics.print(L.FILE .. "   " .. L.EDIT .. "   " .. L.INSERT .. "   " .. L.HELP, 8, 24)
		-- Script display starts at 8, 48 --- now 24 instead of 48
		--for k,v in pairs(scriptlines) do
		
		-- This can roll over, prevent that!
		textlinestogo = 0
		
		-- Display a line for the maximum line size that will fit in VVVVVV!
		love.graphics.line(42*8, 24, 42*8, love.graphics.getHeight())
		
		-- The comment below is a bad way of doing it.
		love.graphics.setScissor(0, 24, love.graphics.getWidth(), love.graphics.getHeight()-24)
		
		-- -- Make sure to display all lines but if we put the cursor further, then do display line numbers
		-- I could make it #scriptlines now
		for k = 1, math.max(table.maxn(scriptlines), editingline) do
			v = anythingbutnil(scriptlines[k])
			
			-- Save the whales, only display this line if we can see it!
			if (scriptscroll+24+(8*k) >= 16) and (scriptscroll+24+(8*k) <= love.graphics.getHeight()) then
				if k >= 500 and editingline == k then
					love.graphics.setColor(255,128,128,255) -- 255 64 64?
				elseif editingline == k then
					love.graphics.setColor(255,255,255,255)
				elseif k >= 500 then
					love.graphics.setColor(255,0,0,255)
				else
					love.graphics.setColor(128,128,128,255)
				end
			
				love.graphics.print(fixdig(k, 3), 8, scriptscroll+24+(8*k))
				textq = syntaxhl(v, 48, scriptscroll+24+(8*k), textlinestogo > 0, editingline == k, syntaxhlon)
			elseif (scriptscroll+24+(8*k) < 16) then
				-- Ok, we could still impact performance if we have TOO MANY say/reply/text commands laying around above this point
				textq = justtext(v, textlinestogo > 0)
			end
			
			if editingline == k then --and textlinestogo == 0 then
				context, carg1, carg2 = scriptcontext(v)
			end
			
			if textq ~= nil then
				textlinestogo = textq
			elseif textlinestogo > 0 then
				textlinestogo = textlinestogo - 1
			end
		end
		
		love.graphics.setScissor()
		
		love.graphics.setColor(255,255,255,255)
		
		-- Now put some buttons on the right!
		--hoverrectangle(128,128,128,128, love.graphics.getWidth()-(128-8), 8+(24*0), 128-16, 16)
		love.graphics.printf(L.FILE, love.graphics.getWidth()-(128-8), 8+(24*0)+4+2, 128-16, "center")
		hoverrectangle(128,128,128,128, love.graphics.getWidth()-(128-8), 8+(24*1), 128-16, 16)
		love.graphics.printf(L.NEW, love.graphics.getWidth()-(128-8), 8+(24*1)+4+2, 128-16, "center")
		hoverrectangle(128,128,128,128, love.graphics.getWidth()-(128-8), 8+(24*2), 128-16, 16)
		love.graphics.printf(L.OPEN, love.graphics.getWidth()-(128-8), 8+(24*2)+4+2, 128-16, "center")
		hoverrectangle(128,128,128,128, love.graphics.getWidth()-(128-8), 8+(24*3), 128-16, 16)
		love.graphics.printf(L.SAVE, love.graphics.getWidth()-(128-8), 8+(24*3)+4+2, 128-16, "center")
		--hoverrectangle(128,128,128,128, love.graphics.getWidth()-(128-8), 8+(24*4), 128-16, 16)
		love.graphics.printf(L.EDITTAB, love.graphics.getWidth()-(128-8), 8+(24*4)+4+2, 128-16, "center")
		hoverrectangle(128,128,128,128, love.graphics.getWidth()-(128-8), 8+(24*5), 128-16, 16)
		love.graphics.printf(L.COPYSCRIPT, love.graphics.getWidth()-(128-8), 8+(24*5)+4+2, 128-16, "center")
		hoverrectangle(128,128,128,128, love.graphics.getWidth()-(128-8), 8+(24*6), 128-16, 16)
		love.graphics.printf(L.SEARCHSCRIPT, love.graphics.getWidth()-(128-8), 8+(24*6)+4+2, 128-16, "center")
		hoverrectangle(128,128,128,128, love.graphics.getWidth()-(128-8), 8+(24*7), 128-16, 16)
		love.graphics.printf(L.GOTOLINE, love.graphics.getWidth()-(128-8), 8+(24*7)+4+2, 128-16, "center")
		--hoverrectangle(128,128,128,128, love.graphics.getWidth()-(128-8), 8+(24*8), 128-16, 16)
		love.graphics.printf(L.VIEW, love.graphics.getWidth()-(128-8), 8+(24*8)+4+2, 128-16, "center")
		hoverrectangle(128,128,128,128, love.graphics.getWidth()-(128-8), 8+(24*9), 128-16, 16)
		love.graphics.printf((syntaxhlon and L.SYNTAXHLOFF or L.SYNTAXHLON), love.graphics.getWidth()-(128-8), 8+(24*9)+4+2, 128-16, "center")
		hoverrectangle(128,128,128,128, love.graphics.getWidth()-(128-8), 8+(24*10), 128-16, 16)
		love.graphics.printf(L.TEXTSIZE, love.graphics.getWidth()-(128-8), 8+(24*10)+4+2, 128-16, "center")
		
		hoverrectangle(128,128,128,128, love.graphics.getWidth()-(128-8), love.graphics.getHeight()-(24*1), 128-16, 16)
		love.graphics.printf(L.RETURN, love.graphics.getWidth()-(128-8), (love.graphics.getHeight()-(24*1))+4+2, 128-16, "center")
		
		-- First make these buttons do things
		if nodialog and love.mouse.isDown("l") then
			if mouseon(love.graphics.getWidth()-(128-8), 8+(24*1), 128-16, 16) then
				-- New
				startmultiinput({""})
				dialog.new(L.NEWSCRIPTNAME, L.CREATENEWSCRIPT, 1, 4, 11)
			elseif mouseon(love.graphics.getWidth()-(128-8), 8+(24*2), 128-16, 16) then
				-- Open
				scriptlines[editingline] = input .. input_r
				scripts[scriptname] = scriptlines
				tostate(10, true)
			elseif mouseon(love.graphics.getWidth()-(128-8), 8+(24*3), 128-16, 16) then
				-- Save
				scriptlines[editingline] = input .. input_r
				input = input .. input_r
				input_r = ""
				scripts[scriptname] = scriptlines
			elseif mouseon(love.graphics.getWidth()-(128-8), 8+(24*5), 128-16, 16) then
				-- Copy script
				love.system.setClipboardText(table.concat(scriptlines, "\r\n"))
			elseif mouseon(love.graphics.getWidth()-(128-8), 8+(24*6), 128-16, 16) then
				-- Search
				dialog.new("Search", "", 1, 1, 0)
			elseif mouseon(love.graphics.getWidth()-(128-8), 8+(24*7), 128-16, 16) then
				-- Go to line
				dialog.new("Go to line", "", 1, 1, 0)
			elseif not mousepressed and mouseon(love.graphics.getWidth()-(128-8), 8+(24*9), 128-16, 16) then
				-- Syntax HL
				syntaxhlon = not syntaxhlon
				
				mousepressed = true
			elseif not mousepressed and mouseon(love.graphics.getWidth()-(128-8), 8+(24*10), 128-16, 16) then
				-- Text size
				textsize = not textsize
				
				mousepressed = true
			elseif mouseon(love.graphics.getWidth()-(128-8), love.graphics.getHeight()-(24*1), 128-16, 16) then
				-- Return
				scriptlines[editingline] = anythingbutnil(input) ..  anythingbutnil(input_r)
				scripts[scriptname] = scriptlines
				tostate(10, true)
				
				mousepressed = true
			end
		end

		if context == "script" then
			love.graphics.printf(carg1, love.graphics.getWidth()-(128-8), 8+(24*11)+4+2, 128-16, "center")
			hoverrectangle(128,128,128,128, love.graphics.getWidth()-(128-8), 8+(24*12), 128-16, 16)
			love.graphics.printf((scripts[carg1] == nil and L.CREATE or L.GOTO), love.graphics.getWidth()-(128-8), 8+(24*12)+4+2, 128-16, "center")
			
			if not mousepressed and nodialog and love.mouse.isDown("l") and mouseon(love.graphics.getWidth()-(128-8), 8+(24*12), 128-16, 16) then
				--dialog.new((scripts[carg1] == nil and "Create" or "Go to") .. " script " .. carg1, "", 1, 1, 0)
				if scripts[carg1] == nil then
					-- Create script (but do not immediately go to it as we can just easily click again)
					scripts[carg1] = {""}
					table.insert(scriptnames, carg1)
				else
					-- Go to script- but save the current script first!
					scriptlines[editingline] = input .. input_r
					scripts[scriptname] = scriptlines
					
					-- Time to go.
					scriptname = carg1
					scriptlines = scripts[carg1]
					--bumpscript(carg1)
					-- Even though we're already in this state, this will initialize stuff.
					tostate(3)
				end
				
				mousepressed = true
			end
		elseif context == "flagscript" then
			love.graphics.printf("Flag " .. carg1, love.graphics.getWidth()-(128-8), 8+(24*11)+4+2, 128-16, "center")
			hoverrectangle(128,128,128,128, love.graphics.getWidth()-(128-8), 8+(24*12), 128-16, 16)
			love.graphics.printf("Name", love.graphics.getWidth()-(128-8), 8+(24*12)+4+2, 128-16, "center")
			
			if not mousepressed and nodialog and love.mouse.isDown("l") and mouseon(love.graphics.getWidth()-(128-8), 8+(24*12), 128-16, 16) then
				dialog.new("Need to make that", "", 1, 1, 0)
				
				mousepressed = true
			end
			if carg2 ~= nil and carg2 ~= "" then
				love.graphics.printf(carg2, love.graphics.getWidth()-(128-8), 8+(24*13)+4+2, 128-16, "center")
				hoverrectangle(128,128,128,128, love.graphics.getWidth()-(128-8), 8+(24*14), 128-16, 16)
				love.graphics.printf((scripts[carg2] == nil and L.CREATE or L.GOTO), love.graphics.getWidth()-(128-8), 8+(24*14)+4+2, 128-16, "center")
				if not mousepressed and nodialog and love.mouse.isDown("l") and mouseon(love.graphics.getWidth()-(128-8), 8+(24*14), 128-16, 16) then
					--dialog.new((scripts[carg2] == nil and "Create" or "Go to") .. " script " .. carg2, "", 1, 1, 0)
					if scripts[carg2] == nil then
						-- Create script (but do not immediately go to it as we can just easily click again)
						scripts[carg2] = {""}
						table.insert(scriptnames, carg2)
					else
						-- Go to script- but save the current script first!
						scriptlines[editingline] = input .. input_r
						scripts[scriptname] = scriptlines
						
						-- Time to go.
						scriptname = carg2
						scriptlines = scripts[carg2]
						--bumpscript(carg2)
						-- Even though we're already in this state, this will initialize stuff.
						tostate(3)
					end
					
					mousepressed = true
				end
			end
		end
	elseif state == 4 then
		love.graphics.print(metadata.Creator, 10, 10)
		love.graphics.print("That should say Unknown.", 10, 20)
	elseif state == 5 then
		love.graphics.print("Userprofile: " .. userprofile, 8, 8)
		for k,v in pairs(files) do
			love.graphics.print(v, 8, 16+8*k)
			
			lastk = k
		end
				
		love.graphics.print("Levels folder error: " .. lerror .. " (0 means no error)", 8, 16+8*lastk+16)
		love.graphics.print("Levels folder: " .. anythingbutnil(levelsfolder), 8, 16+8*lastk+24)
	elseif state == 6 then
		love.graphics.print(L.LEVELSLIST, 8, 8)
		
		if lerror ~= 0 then
			love.graphics.printf(langkeys(L.COULDNOTGETCONTENTSLEVELFOLDER, {levelsfolder}), 8, 24, love.graphics.getWidth()-16, "left")
		else
			local k2 = 1
			for k,v in pairs(files) do
				if input .. input_r == "" or v:lower():find("^" .. escapegsub(input .. input_r)) then
					if tabselected == k2 then
						love.graphics.setColor(255,255,0)
					end
					
					love.graphics.print(v, 8, 16+8*k2) -- y = 16+8*k
					
					love.graphics.setColor(255,255,255)
					
					--[[
					if k2 == 1 and love.keyboard.isDown("tab") then
						input = v:sub(1, -8)
					end
					]]
					if tabselected == k2 and love.keyboard.isDown("return") then
						state6load(v:sub(1, -8))
					end
					
					k2 = k2 + 1
				end
					
				lastk = k
			end
			
			if tabselected == 0 and (love.keyboard.isDown("up") or ((love.keyboard.isDown("lshift") or love.keyboard.isDown("rshift")) and love.keyboard.isDown("tab"))) then
				-- Start from the bottom.
				tabselected = k2-1
			elseif tabselected >= k2 then
				tabselected = 0
			end
		end
		
		love.graphics.print(L.LOADTHISLEVEL .. input .. __, 10, love.graphics.getHeight()-18)
		startinputonce()
		
		hoverdraw(helpbtn, love.graphics.getWidth()-128+8, 8, 16, 16, 1)
		hoverdraw(newbtn, love.graphics.getWidth()-32, 0, 32, 32, 2) -- -96
		
		if not mousepressed and nodialog and love.mouse.isDown("l") then
			if mouseon(love.graphics.getWidth()-128+8, 8, 16, 16) then
				tostate(15)
			elseif mouseon(love.graphics.getWidth()-32, 0, 32, 32) then -- -96
				-- New
				stopinput()
				success, metadata, roomdata, entitydata, levelmetadata, scripts, count, scriptnames, vedmetadata = createblanklevel(5, 5)
				editingmap = "untitled\n"
				tostate(1)
				-- Make sure we don't immediately click the next button
				nodialog = false
				mousepressed = true
			end
		end

		--love.graphics.print("Levels folder error: " .. lerror .. " (0 means no error)", 8, 16+8*lastk+16)
		--love.graphics.print("Levels folder: " .. anythingbutnil(levelsfolder), 8, 16+8*lastk+24)
		
		hoverrectangle(128,128,128,128, love.graphics.getWidth()-(128-8), 48-8, 128-16, 16)
		love.graphics.printf(L.VEDOPTIONS, love.graphics.getWidth()-(128-8), (48-8)+4+2, 128-16, "center")
		hoverrectangle(128,128,128,128, love.graphics.getWidth()-(128-8), 40+120, 128-16, 16)
		love.graphics.printf(L.SENDFEEDBACK, love.graphics.getWidth()-(128-8), (40+120)+4+2, 128-16, "center") -- Test
		
		versionchecked = verchannel:peek()
		
		if versionchecked ~= nil then
			if versionchecked == "connecterror" or versionchecked == "error" then
				love.graphics.printf(L.VERSIONERROR, love.graphics.getWidth()-(128-8), 40+120+16+3+8, 128-16, "left")
			elseif versionchecked == "latest" then
				love.graphics.printf(L.VERSIONUPTODATE, love.graphics.getWidth()-(128-8), 40+120+16+3+8, 128-16, "left")
			else
				love.graphics.printf(langkeys(L.VERSIONOLD, {versionchecked}), love.graphics.getWidth()-(128-8), 40+120+16+3+8, 128-16, "left")
				updatebutton = true
			end
		else
			love.graphics.printf(L.VERSIONCHECKING, love.graphics.getWidth()-(128-8), 40+120+16+3+8, 128-16, "left")
		end
		
		if not mousepressed and nodialog and love.mouse.isDown("l") then
			if mouseon(love.graphics.getWidth()-(128-8), 48-8, 128-16, 16) then
				-- Ved options
				--dialog.new("Ved options coming soon\n\nFor now, you can edit the config file at %appdata%\\LOVE\\ved\\settings.lua (Windows) or Library/Application Support/LOVE/ved/settings.lua (Mac)\n\nThis dialog is so temporary that I won't bother checking the OS for it to display the correct folder name.", "", 1, 1, 0)
				tostate(13)
				
				mousepressed = true
			elseif not mousepressed and mouseon(love.graphics.getWidth()-(128-8), 40+120, 128-16, 16) then
				-- Test BUT "SEND FEEDBACK" FOR NOW
				--dialog.new("Auto-creation of a save file for VVVVVV coming soon!", "", 1, 1, 0)
				openurl("http://ved.idea.informer.com/")
				
				mousepressed = true
			end
		end
	elseif state == 7 then
		for y = 0, 7 do
			for x = 0, 23 do
				love.graphics.draw(tilesets["sprites.png"]["img"], tilesets["sprites.png"]["tiles"][(y*24)+x], x*32, y*32)
				
				if mouseon(x*32, y*32, 32, 32) then
					love.graphics.print((y*24)+x, love.graphics.getWidth()-24, love.graphics.getHeight()-8)
				end
			end
		end
	elseif state == 8 then
		love.graphics.print(L.ENTERNAMESAVE .. input .. __, 10, 10)
		startinputonce()
		if savedsuccess ~= nil then
			if savedsuccess == true then
				love.graphics.print(L.SAVESUCCESS, 10, 50)
			else
				love.graphics.print(L.SAVENOSUCCESS .. anythingbutnil(savederror), 10, 50)
			end
		end
	elseif state == 9 then
		love.graphics.print(youdidanswer .. "\nRight click menu return: " .. anythingbutnil(RCMreturn), 10, 10)
	elseif state == 10 then
		--[[
		j = -1
		for k,v in pairs(scripts) do
			j = j + 1
			hoverrectangle(128,128,128,128, 8, 8+(24*j), 640, 16)
			love.graphics.printf(k, 8, 8+(24*j)+4+2, 640, "center")
		end
		]]
		j = -1
		for rvnum = #scriptnames, 1, -1 do
			j = j + 1
			hoverrectangle(128,128,128,128, 8, scriptlistscroll+8+(24*j), 64+64+640-8, 16)
			love.graphics.printf(scriptnames[rvnum], 8, scriptlistscroll+8+(24*j)+4+2, 64+64+640-8, "center")
			
			-- Are we clicking on this?
			if nodialog and mouseon(8, scriptlistscroll+8+(24*j), 64+64+640-8, 16) then
				if love.mouse.isDown("l") then
					scriptname = scriptnames[rvnum]
					scriptlines = scripts[scriptnames[rvnum]]
					processflaglabels()
					bumpscript(rvnum)
					tostate(3)
				elseif love.mouse.isDown("r") then
					rightclickmenu.create({L.EDIT, L.EDITWOBUMPING, L.COPYNAME, L.COPYCONTENTS, L.DUPLICATE, L.DELETE}, "spt_" .. rvnum)
				end
			end
		end

		hoverrectangle(128,128,128,128, love.graphics.getWidth()-(128-8), 8+(24*0), 128-16, 16)
		love.graphics.printf(L.NEW, love.graphics.getWidth()-(128-8), 8+(24*0)+4+2, 128-16, "center")
		
		hoverrectangle(128,128,128,128, love.graphics.getWidth()-(128-8), love.graphics.getHeight()-(24*1), 128-16, 16)
		love.graphics.printf(L.RETURN, love.graphics.getWidth()-(128-8), (love.graphics.getHeight()-(24*1))+4+2, 128-16, "center")
		
		-- Buttons again
		if nodialog and love.mouse.isDown("l") then
			if mouseon(love.graphics.getWidth()-(128-8), 8+(24*0), 128-16, 16) then
				-- New
				--dialog.new("New", "", 1, 1, 0)
				startmultiinput({""})
				dialog.new(L.NEWSCRIPTNAME, L.CREATENEWSCRIPT, 1, 4, 9)
			elseif not mousepressed and mouseon(love.graphics.getWidth()-(128-8), love.graphics.getHeight()-(24*1), 128-16, 16) then
				-- Return
				tostate(1, true)
			end
		end
	elseif state == 11 then
		love.graphics.print(L.SEARCHFOR .. input .. __, 5, 5) -- Temporary location and temporary untextbox
		
		for k,v in pairs(searchscripts) do
			if k <= showresults then
				love.graphics.setScissor(8, 32*k, 288, 30)
				hoverrectangle(128,128,128,128, 8, 32*k, 288, 30)
				--love.graphics.print(v.name, 12, 32*k+7)
				highlightresult(v.name, searchedfor, 12, 32*k+7)
				if v.foundline ~= 0 then
					--love.graphics.print(v.foundline .. ": " .. v.foundlinecontent, 12, 32*k+8+7)
					love.graphics.print(", " .. L.LINE .. v.foundline, 12+v.name:len()*8, 32*k+7)
					highlightresult(v.foundlinecontent, searchedfor, 12, 32*k+8+7)
				end
			
				if nodialog and love.mouse.isDown("l") and mouseon(8, 32*k, 288, 30) then
					scriptname = v.name
					scriptlines = scripts[v.name]
					processflaglabels()
					--bumpscript(rvnum) do bump later though
					tostate(3)
					
					if v.foundline ~= 0 then
						editingline = v.foundline
					end
					
					mousepressed = true
					nodialog = false
				end
			end
		end
		
		for k,v in pairs(searchrooms) do
			if k <= showresults then
				love.graphics.setScissor(8+288+8, 32*k, 288, 30)
				hoverrectangle(128,128,128,128, 8+288+8, 32*k, 288, 30)
				love.graphics.print("(" .. v.x .. "," .. v.y .. ")", 12+288+8, 32*k+7)
				highlightresult(v.name, searchedfor, 12+288+8, 32*k+8+7)
				
				if nodialog and love.mouse.isDown("l") and mouseon(8+288+8, 32*k, 288, 30) then
					roomx = v.x - 1
					roomy = v.y - 1
					
					selectedtileset = levelmetadata[(roomy)*20 + (roomx+1)].tileset
					selectedcolor = levelmetadata[(roomy)*20 + (roomx+1)].tilecol
		
					tostate(1, true)
					
					mousepressed = true
					nodialog = false
				end
			end
		end
		
		love.graphics.setScissor()
		
		--love.graphics.rectangle("fill", 8, 32, 288, 32)
		--love.graphics.rectangle("fill", 8+288+8, 32, 288, 32)
		--love.graphics.rectangle("fill", 8+288+8+288+8, 32, 288, 32)
	elseif state == 12 then
		if mapscreenshot == nil then
			-- Display this just once and then cache that, we don't like to have 1 FPS
			for mry = 0, metadata.mapheight-1 do
				for mrx = 0, metadata.mapwidth-1 do
					displayroom(mapxoffset+64+64+(mrx*mapscale*640), mapyoffset+mry*mapscale*480, roomdata[mry][mrx], levelmetadata[(mry)*20 + (mrx+1)], mapscale)
				end
			end
			
			mapscreenshot = love.graphics.newImage(love.graphics.newScreenshot())
		else
			love.graphics.draw(mapscreenshot, 0, 0, 0, s.pscale^-1)
			
			-- We should be able to hover over rooms
			hoverx = nil
			hovery = nil
			hovername = nil
			
			for mry = 0, metadata.mapheight-1 do
				for mrx = 0, metadata.mapwidth-1 do
					if mouseon(mapxoffset+64+64+(mrx*mapscale*640), mapyoffset+mry*mapscale*480, mapscale*640, mapscale*480) then
						love.graphics.setColor(255,255,255,64)
						love.graphics.rectangle("fill", mapxoffset+64+64+(mrx*mapscale*640), mapyoffset+mry*mapscale*480, mapscale*640, mapscale*480)
						love.graphics.setColor(255,255,255,255)
						
						hoverx = mrx
						hovery = mry
						hovername = levelmetadata[(mry)*20 + (mrx+1)].roomname
						
						-- But maybe we're clicking this room!
						if love.mouse.isDown("l") then
							-- Is this an action?
							if not mousepressed and selectingrooms == 0 then
								-- Nope
								roomx = hoverx
								roomy = hovery
								
								selectedtileset = levelmetadata[(roomy)*20 + (roomx+1)].tileset
								selectedcolor = levelmetadata[(roomy)*20 + (roomx+1)].tilecol
								
								-- We don't want to click the first tile we press
								nodialog = false
								
								tostate(1, true)
							elseif not mousepressed and selected1x == -1 then
								-- Select 1
								selected1x = hoverx
								selected1y = hovery
								mousepressed = true
							elseif not mousepressed and selected2x == -1 then
								-- Select 2 and proceed
								selected2x = hoverx
								selected2y = hovery
								
								-- What were we selecting the rooms for, btw?
								if selectingrooms == 1 then
									-- Copying!
									cons("Copying room data...")
									--roomdata[selected2y][selected2x] = roomdata[selected1y][selected1x]
									local copieddata = getroomcopydata(selected1x, selected1y)
									setroomfromcopy(copieddata, selected2x, selected2y)
									temporaryroomnametimer = 0
									cons("Copying entities...")
									copymoveentities(selected1x, selected1y, selected2x, selected2y, false)
									cons("Done...")
									
									-- Refresh the map.
									mapscreenshot = nil
									collectgarbage("collect")
									
									-- Hey wait don't create a COMPLETELY new map
									mrx = selected2x
									mry = selected2y
									love.graphics.setColor(0,0,0,255)
									love.graphics.rectangle("fill", mapxoffset+64+64+(mrx*mapscale*640), mapyoffset+mry*mapscale*480, mapscale*640, mapscale*480)
									love.graphics.setColor(255,255,255,255)
									displayroom(mapxoffset+64+64+(mrx*mapscale*640), mapyoffset+mry*mapscale*480, roomdata[mry][mrx], levelmetadata[(mry)*20 + (mrx+1)], mapscale)
									
									mapscreenshot = love.graphics.newImage(love.graphics.newScreenshot())
								elseif selectingrooms == 2 then
									-- Swapping!
									
								end
								
								selectingrooms = 0
								mousepressed = true
							end
						end
					end
				end
			end
			
			if (hoverx ~= nil) and (hovery ~= nil) then
				love.graphics.print("(" .. (hoverx+1) .. "," .. (hovery+1) .. ")", 64+64+640, 3)
			end
			if (hovername ~= nil) then
				love.graphics.printf(hovername, 64+64+640, 11, love.graphics.getWidth()-64-64-640, "left")
			end
			
			hoverrectangle(128,128,128,128, love.graphics.getWidth()-(128-8), love.graphics.getHeight()-(24*1), 128-16, 16)
			love.graphics.printf(L.RETURN, love.graphics.getWidth()-(128-8), (love.graphics.getHeight()-(24*1))+4+2, 128-16, "center")
			hoverrectangle(128,128,128,128, love.graphics.getWidth()-(128-8), love.graphics.getHeight()-(24*2), 128-16, 16)
			love.graphics.printf(L.SAVEMAP, love.graphics.getWidth()-(128-8), (love.graphics.getHeight()-(24*2))+4+2, 128-16, "center")
			hoverrectangle(128,128,128,128, love.graphics.getWidth()-(128-8), love.graphics.getHeight()-(24*4), 128-16, 16)
			love.graphics.printf(L.COPYROOMS, love.graphics.getWidth()-(128-8), (love.graphics.getHeight()-(24*4))+4+2, 128-16, "center")
			hoverrectangle(128,128,128,128, love.graphics.getWidth()-(128-8), love.graphics.getHeight()-(24*5), 128-16, 16)
			love.graphics.printf(L.SWAPROOMS, love.graphics.getWidth()-(128-8), (love.graphics.getHeight()-(24*5))+4+2, 128-16, "center")
			
			-- The buttons are clickable
			if nodialog and love.mouse.isDown("l") then
				if mouseon(love.graphics.getWidth()-(128-8), love.graphics.getHeight()-(24*1), 128-16, 16) then
					-- Return
					tostate(1, true)
				elseif mouseon(love.graphics.getWidth()-(128-8), love.graphics.getHeight()-(24*2), 128-16, 16) then
					-- Save map
					saveas = (editingmap == "untitled\n" and "untitled" or editingmap) .. "_" .. os.time() .. ".png"
					mapscreenshot:getData():encode("maps/" .. saveas)
					dialog.new(langkeys(L.MAPSAVEDAS, {saveas}), "", 1, 1, 0)
				elseif mouseon(love.graphics.getWidth()-(128-8), love.graphics.getHeight()-(24*4), 128-16, 16) then
					-- Copy rooms
					--dialog.new("Coming soon!", "", 1, 1, 0)
					selectingrooms = 1
					selected1x = -1; selected1y = -1
					selected2x = -1; selected2y = -1
				elseif mouseon(love.graphics.getWidth()-(128-8), love.graphics.getHeight()-(24*5), 128-16, 16) then
					-- Swap rooms
					dialog.new("Coming soon!", "", 1, 1, 0)
				end
			elseif nodialog and love.mouse.isDown("r") then
				if mouseon(love.graphics.getWidth()-(128-8), love.graphics.getHeight()-(24*2), 128-16, 16) then
					-- Save map
					rightclickmenu.create({L.SAVEMAP, L.SAVEFULLSIZEMAP}, "savemap")
				end
			end
		end
	elseif state == 13 then
		-- Options screen
		--love.graphics.draw(checkon, love.graphics.getWidth()-98, 50, 0, 2)
		--love.graphics.draw(checkoff, love.graphics.getWidth()-98, 70, 0, 2)
		love.graphics.print(L.VEDOPTIONS, 8, 8+4+2)
		
		love.graphics.print(L.CUSTOMVVVVVVDIRECTORY, 8+16+8, 8+(24*1)+4+2)
		
		hoverdraw((s.scale == 2 and checkon or checkoff), 8, 8+(24*2), 16, 16, 2)
		love.graphics.print("2x scale", 8+16+8, 8+(24*2)+4+2)
		
		hoverdraw((s.dialoganimations and checkon or checkoff), 8, 8+(24*3), 16, 16, 2)
		love.graphics.print(L.DIALOGANIMATIONS, 8+16+8, 8+(24*3)+4+2)
		
		hoverdraw((s.flipsubtoolscroll and checkon or checkoff), 8, 8+(24*4), 16, 16, 2)
		love.graphics.print(L.FLIPSUBTOOLSCROLL, 8+16+8, 8+(24*4)+4+2)
		
		hoverdraw((s.adjacentroomlines and checkon or checkoff), 8, 8+(24*5), 16, 16, 2)
		love.graphics.print(L.ADJACENTROOMLINES, 8+16+8, 8+(24*5)+4+2)
		
		hoverdraw((s.askbeforequit and checkon or checkoff), 8, 8+(24*6), 16, 16, 2)
		love.graphics.print(L.ASKBEFOREQUIT, 8+16+8, 8+(24*6)+4+2)
		
		
		hoverrectangle(128,128,128,128, love.graphics.getWidth()-(128-8), 8+(24*0), 128-16, 16)
		love.graphics.printf("OK", love.graphics.getWidth()-(128-8), 8+(24*0)+4+2, 128-16, "center")
		
		--hoverrectangle(128,128,128,128, love.graphics.getWidth()-(128-8), 8+(24*1), 128-16, 16) -- love.graphics.getHeight()-(24*(X+1)) instead of 8+(24*X)
		--love.graphics.printf("Cancel", love.graphics.getWidth()-(128-8), (8+(24*1))+4+2, 128-16, "center")
		
		
		if nodialog and not mousepressed and love.mouse.isDown("l") then
			if mouseon(8, 8+(24*2), 16, 16) then
				-- 2x scale
				if s.scale == 2 then
					s.scale = 1
				else
					s.scale = 2
				end
				dialog.new("The new scale settings will take effect after rebooting Ved.", "", 1, 1, 0)
			elseif mouseon(8, 8+(24*3), 16, 16) then
				-- Dialog animations
				s.dialoganimations = not s.dialoganimations
			elseif mouseon(8, 8+(24*4), 16, 16) then
				-- Flip subtool scrolling direction
				s.flipsubtoolscroll = not s.flipsubtoolscroll
			elseif mouseon(8, 8+(24*5), 16, 16) then
				-- Indicators of tiles in adjacent rooms
				s.adjacentroomlines = not s.adjacentroomlines
			elseif mouseon(8, 8+(24*6), 16, 16) then
				-- Ask before quitting
				s.askbeforequit = not s.askbeforequit
				
			elseif mouseon(love.graphics.getWidth()-(128-8), 8+(24*0), 128-16, 16) then
				-- Save
				saveconfig()
				tostate(oldstate)
			--elseif not mousepressed and mouseon(love.graphics.getWidth()-(128-8), 8+(24*1), 128-16, 16) then
				-- Cancel
				--tostate(oldstate)
			end
			
			mousepressed = true
		end
	elseif state == 14 then
		for rrrrr = 0, 1 do
			for asdfg = 0, 4 do
				drawentity(enemysprites[5*rrrrr+asdfg], 16+48*asdfg, 16+48*rrrrr)
			end
		end
	elseif state == 15 then
		-- Leaving room for a 16 px wide scrollbar. 4+16+8 
		
		j = -1
		for rvnum = 1, #helppages+(helpeditable and 1 or 0) do
			j = j + 1
			if helpeditingline ~= 0 then
				love.graphics.setColor(128,128,128,64)
				love.graphics.rectangle("fill", 8, helplistscroll+8+(24*j), 25*8-28, 16)
				love.graphics.setColor(255,255,255,128)
			else
				hoverrectangle(128,128,128,128, 8, helplistscroll+8+(24*j), 25*8-28, 16)
			end
			love.graphics.printf((helppages[rvnum] == nil and "[Add new]" or helppages[rvnum].subj), 8, helplistscroll+8+(24*j)+4+2, 25*8-28, "center")
			
			-- Are we clicking on this?
			if nodialog and helpeditingline == 0 and mouseon(8, helplistscroll+8+(24*j), 25*8-28, 16) then
				if love.mouse.isDown("l") then
					if helppages[rvnum] == nil then
						-- This is just the "add new" button.
						startmultiinput({""})
						dialog.new(L.NEWNOTENAME, L.CREATENEWNOTE, 1, 4, 12)
					else
						helparticle = rvnum
						helparticlecontent = explode("\n", helppages[helparticle].cont)
						helparticlescroll = 0
					end
				end
			end
		end
		
		-- Scroll bar for list of articles
		--if scrollbar(8+(25*8-28)+4, 8, love.graphics.getHeight()-16, (love.graphics.getHeight()-16)*2, 0) then
		
		--end
		
		local linee = 0
		
		if helppages[helparticle] ~= nil then
			-- ...
			--sis = explode("\n", helppages[helparticle].cont)
			
			--for s in string.gmatch(LH[helparticle].cont, ".*\n") do
			love.graphics.setScissor(8+200+8-2, 8, love.graphics.getWidth()-(8+200+8-4), love.graphics.getHeight()-16)
			love.graphics.setColor(192,192,192,255)

			for k,s in pairs(helparticlecontent) do
				if helpeditingline == k then
					love.graphics.print(s .. __, 8+200+8, helparticlescroll+8+(10*linee)+4+2)
				elseif s:find("\\") then
					local imageshift = 0
					local imagex = 0
					
					part1, part2 = s:match("(.*)\\(.*)")
					
					for fl = 1, part2:len() do
						if part2:sub(fl,fl) == "h" then
							love.graphics.setFont(font16)
						elseif part2:sub(fl,fl) == "#" then
							
						elseif tonumber(part2:sub(fl,fl)) ~= nil then
							local dodisplay = true
						
							if helpimages[helparticle .. "_" .. (imageshift+tonumber(part2:sub(fl,fl)))] == nil and helppages[helparticle].imgs[imageshift+tonumber(part2:sub(fl,fl))+1] ~= nil then
								if love.filesystem.exists(helppages[helparticle].imgs[imageshift+tonumber(part2:sub(fl,fl))+1]) then
									cons("Loading help image " .. helparticle .. "_" .. (imageshift+tonumber(part2:sub(fl,fl))))
									helpimages[helparticle .. "_" .. (imageshift+tonumber(part2:sub(fl,fl)))] = love.graphics.newImage(helppages[helparticle].imgs[imageshift+tonumber(part2:sub(fl,fl))+1])
								else
									dodisplay = false
									love.graphics.setColor(255,0,0,255)
									part1 = "[IMAGE ERROR]"
								end
							elseif helppages[helparticle].imgs[imageshift+tonumber(part2:sub(fl,fl))+1] == nil then
								dodisplay = false
								love.graphics.setColor(255,0,0,255)
								part1 = "[IMAGE ERROR]"
							end
							
							if dodisplay then
								love.graphics.draw(helpimages[helparticle .. "_" .. (imageshift+tonumber(part2:sub(fl,fl)))], 8+200+8+imagex, helparticlescroll+8+(10*linee)+4)
							end
						elseif part2:sub(fl,fl) == "^" then
							imageshift = imageshift + 10
						elseif part2:sub(fl,fl) == "_" then
							imageshift = imageshift - 10
						elseif part2:sub(fl,fl) == ">" then
							imagex = imagex + 8
						elseif part2:sub(fl,fl) == "<" then
							imagex = imagex - 8
						elseif part2:sub(fl,fl) == "r" then
							love.graphics.setColor(255,0,0,255)
						elseif part2:sub(fl,fl) == "g" then
							love.graphics.setColor(128,128,128,255)
						elseif part2:sub(fl,fl) == "w" then
							love.graphics.setColor(255,255,255,255)
						elseif part2:sub(fl,fl) == "b" then
							love.graphics.setColor(64,64,255,255)
						elseif part2:sub(fl,fl) == "o" then
							love.graphics.setColor(255,128,0,255)
						elseif part2:sub(fl,fl) == "v" then
							love.graphics.setColor(0,255,0,255)
						
						elseif part2:sub(fl,fl) == "C" then
							love.graphics.setColor(132, 181, 255)
						elseif part2:sub(fl,fl) == "P" then
							love.graphics.setColor(255, 135, 255)
						elseif part2:sub(fl,fl) == "Y" then
							love.graphics.setColor(255, 255, 135)
						elseif part2:sub(fl,fl) == "R" then
							love.graphics.setColor(255, 61, 61)
						elseif part2:sub(fl,fl) == "G" then
							love.graphics.setColor(144, 255, 144)
						elseif part2:sub(fl,fl) == "B" then
							love.graphics.setColor(75, 75, 230)
						elseif part2:sub(fl,fl) == "-" then
							part1 = ("_"):rep(82)
						elseif part2:sub(fl,fl) == ")" then
							tostate(oldstate)
							nodialog = false
						end
					end
					love.graphics.print(part1, 8+200+8, helparticlescroll+8+(10*linee)+4+2)
					love.graphics.setFont(font8)
					love.graphics.setColor(192,192,192,255)
				else
					love.graphics.print(s, 8+200+8, helparticlescroll+8+(10*linee)+4+2)
				end
				
				linee = linee + 1
			end
			love.graphics.setColor(255,255,255,255)
			love.graphics.setScissor()
			
			-- Scroll bar for the article itself
			if scrollbar(love.graphics.getWidth()-24, 8, love.graphics.getHeight()-16, (#helparticlecontent*10)-love.graphics.getHeight()-16, ((-helparticlescroll))/(#helparticlecontent*10-love.graphics.getHeight()-16)) then
			
			end
			
			-- Is this note editable?
			if helpeditable then
				-- It is!
				if helpeditingline == 0 then
					-- We're not currently editing it.
					hoverrectangle(128,128,128,128, love.graphics.getWidth()-(128-8)-16-4-(128-12)-(128-12), love.graphics.getHeight()-(24*1), 128-16, 16)
					love.graphics.printf("Rename", love.graphics.getWidth()-(128-8)-16-4-(128-12)-(128-12), (love.graphics.getHeight()-(24*1))+4+2, 128-16, "center")
					hoverrectangle(128,128,128,128, love.graphics.getWidth()-(128-8)-16-4-(128-12), love.graphics.getHeight()-(24*1), 128-16, 16)
					love.graphics.printf(L.EDIT, love.graphics.getWidth()-(128-8)-16-4-(128-12), (love.graphics.getHeight()-(24*1))+4+2, 128-16, "center")
					hoverrectangle(128,128,128,128, love.graphics.getWidth()-(128-8)-16-4, love.graphics.getHeight()-(24*1), 128-16, 16)
					love.graphics.printf(L.DELETE, love.graphics.getWidth()-(128-8)-16-4, (love.graphics.getHeight()-(24*1))+4+2, 128-16, "center")
					
					if nodialog and love.mouse.isDown("l") then
						if mouseon(love.graphics.getWidth()-(128-8)-16-4-(128-12)-(128-12), love.graphics.getHeight()-(24*1), 128-16, 16) then
							-- Rename
							startmultiinput({helppages[helparticle].subj})
							dialog.new("New name:", "Rename note", 1, 4, 13)
						elseif mouseon(love.graphics.getWidth()-(128-8)-16-4-(128-12), love.graphics.getHeight()-(24*1), 128-16, 16) then
							-- Edit
							helpeditingline = 1
							input = anythingbutnil(helparticlecontent[helpeditingline])
							takinginput = true
							nodialog = false
						elseif mouseon(love.graphics.getWidth()-(128-8)-16-4, love.graphics.getHeight()-(24*1), 128-16, 16) then
							-- Delete
							dialog.new("Are you sure you want to delete this note?", "", 1, 3, 14)
						end
					end
				else
					-- We are currently editing it.
					hoverrectangle(128,128,128,128, love.graphics.getWidth()-(128-8)-16-4, love.graphics.getHeight()-(24*1), 128-16, 16)
					love.graphics.printf(L.SAVE, love.graphics.getWidth()-(128-8)-16-4, (love.graphics.getHeight()-(24*1))+4+2, 128-16, "center")
					
					if nodialog and love.mouse.isDown("l") then
						if mouseon(love.graphics.getWidth()-(128-8)-16-4, love.graphics.getHeight()-(24*1), 128-16, 16) then
							-- Save
							helparticlecontent[helpeditingline] = input .. input_r
							helpeditingline = 0
							stopinput()
							takinginput = false
							helppages[helparticle].cont = table.concat(helparticlecontent, "\n")
							nodialog = false
						end
					end
				end
			end
		end
	elseif state == 16 then
		
	elseif state == 17 then
		love.graphics.print(love.filesystem.read("folderopendialog_return.txt"), 8, 8)
	else
		fatalerror(langkeys(L.UNKNOWNSTATE, {state, oldstate}))
	end
	
	rightclickmenu.draw()
	
	dialog.draw()
	love.graphics.setColor(255,255,255,255)
	
	-- Show debug code if wanted
	if allowdebug then
		
	end
	
	if fpscap == 1 then
		love.graphics.setColor(255,0,0)
		love.graphics.setFont(font16)
		love.graphics.print("60", love.graphics.getWidth()-32, love.graphics.getHeight()-16)
		love.graphics.setFont(font8)
		love.graphics.setColor(255,255,255)
	elseif fpscap == 2 then
		love.graphics.setColor(255,0,0)
		love.graphics.setFont(font16)
		love.graphics.print("30", love.graphics.getWidth()-32, love.graphics.getHeight()-16)
		love.graphics.setFont(font8)
		love.graphics.setColor(255,255,255)
	elseif fpscap == 3 then
		love.graphics.setColor(255,0,0)
		love.graphics.setFont(font16)
		love.graphics.print("15", love.graphics.getWidth()-32, love.graphics.getHeight()-16)
		love.graphics.setFont(font8)
		love.graphics.setColor(255,255,255)
	end
	
	
	if s.pscale ~= 1 then
		love.graphics.pop()
	end
end

function love.update(dt)
	if takinginput then
		cursorflashtime = (cursorflashtime + dt) % 1
		--__ = (cursorflashtime <= .5 and "_" or (input_r:sub(1, 1) == "" and " " or firstUTF8(input_r))) .. input_r:sub(2, -1)
		firstchar = firstUTF8(input_r)
		if cursorflashtime <= .5 then
			__ = "_" .. input_r:sub(1 + firstchar:len())
		else
			__ = firstchar .. input_r:sub(1 + firstchar:len())
		end
	end
	
	if fpscap == 1 then
		if dt < 1/60 then
			love.timer.sleep(1/60 - dt)
		end
	elseif fpscap == 2 then
		if dt < 1/30 then
			love.timer.sleep(1/30 - dt)
		end
	elseif fpscap == 3 then
		if dt < 1/15 then
			love.timer.sleep(1/15 - dt)
		end
	end
	
	-- The timing for this doesn't really matter
	if temporaryroomnametimer > 0 then
		temporaryroomnametimer = temporaryroomnametimer - 1
	end

	if state ~= -1 then
		if allowdebug then
			love.window.setTitle(anythingbutnil(editingmap) .. " - Ved α" .. ver .. "  [Debug mode on]  [" .. L.FPS .. ": " .. love.timer.getFPS() .. "] - " .. L.STATE .. ": " .. state .. " - " .. love.graphics.getWidth() .. "x" .. love.graphics.getHeight() .. " " .. L.MOUSE .. ": " .. love.mouse.getX() .. " " .. love.mouse.getY() .. "  [ LÖVE v0." .. (love.graphics.ellipse == nil and 9 or 10) .. " ]")
		else
			love.window.setTitle(anythingbutnil(editingmap) .. " - Ved α" .. ver .. "  [" .. L.FPS .. ": " .. love.timer.getFPS() .. "]")
		end
	end
	
	if RCMactive or (DIAwindowani ~= 16) then
		nodialog = false
	elseif not love.mouse.isDown("l") then
		nodialog = true
	end
	
	-- Right click menu
	if RCMreturn ~= nil and RCMreturn ~= "" then
		if RCMid:sub(1, 4) == "ent_" then
			-- Something to do with an entity.
			entdetails = explode("_", RCMid)
			if entitydata[tonumber(entdetails[3])] ~= nil then
				if RCMreturn == L.DELETE then
					-- Delete this entity, whatever it is. But update the counter as well! And if we were editing that roomtext, stop
					if (tonumber(entdetails[2]) == 17 or tonumber(entdetails[2] == 18) or tonumber(entdetails[2] == 19)) and editingroomtext == tonumber(entdetails[3]) then
						stopinput()
						editingroomtext = 0
						makescriptroomtext = false
					elseif tonumber(entdetails[2]) == 16 then
						count.startpoint = nil
					end
					updatecountdelete(tonumber(entdetails[2]))
					entitydata[tonumber(entdetails[3])] = nil
					count.entities = count.entities - 1
				elseif RCMreturn == L.PROPERTIES then
					-- Edit properties of this entity, whatever it is. But if we were editing room text or a name of something, stop that first.
					if editingroomtext > 0 then
						-- The argument here is the number of the entity not to make nil- the entity we currently want to edit the properties of
						endeditingroomtext(tonumber(entdetails[3]))
					end
					
					thisentity = entitydata[tonumber(entdetails[3])]
					startmultiinput({thisentity.x, thisentity.y, thisentity.t, thisentity.p1, thisentity.p2, thisentity.p3, thisentity.p4, thisentity.p5, thisentity.p6, thisentity.data})
					dialog.new(L.RAWENTITYPROPERTIES .. "\n\nx\ny\nt\np1\np2\np3\np4\np5\np6\ndata", (allowdebug and "[ID: " .. tonumber(entdetails[3]) .. "] (do not rely on the ID)" or ""), 1, 5, 2)
				elseif tonumber(entdetails[2]) == 1 then
					-- Enemy
					if RCMreturn == L.CHANGEDIRECTION then
						entitydata[tonumber(entdetails[3])].p1 = cycle(entitydata[tonumber(entdetails[3])].p1, 3, 0)
					else
						dialog.new(RCMid .. " " .. RCMreturn .. " not supported yet.", "", 1, 1, 0)
					end
				elseif tonumber(entdetails[2]) == 2 then
					-- Platform, moving/conveyor
					if RCMreturn == L.CYCLETYPE then
						if entitydata[tonumber(entdetails[3])].p1 < 4 then
							-- Moving platform
							entitydata[tonumber(entdetails[3])].p1 = cycle(entitydata[tonumber(entdetails[3])].p1, 3, 0)
						else
							-- Conveyor
							entitydata[tonumber(entdetails[3])].p1 = cycle(entitydata[tonumber(entdetails[3])].p1, 8, 5)
						end
					else
						dialog.new(RCMid .. " " .. RCMreturn .. " not supported yet.", "", 1, 1, 0)
					end
				elseif tonumber(entdetails[2]) == 10 then
					-- Checkpoint
					if RCMreturn == L.FLIP then
						entitydata[tonumber(entdetails[3])].p1 = cycle(entitydata[tonumber(entdetails[3])].p1, 1, 0)
					else
						dialog.new(RCMid .. " " .. RCMreturn .. " not supported yet.", "", 1, 1, 0)
					end
				elseif tonumber(entdetails[2]) == 11 then
					-- Gravity line
					--[[
					if RCMreturn == L.CHANGETOHOR then
						entitydata[tonumber(entdetails[3])].p1 = 0
						dialog.new("I know this looks bizarre.", "", 1, 1, 0)
					elseif RCMreturn == L.CHANGETOVER then
						entitydata[tonumber(entdetails[3])].p1 = 1
						dialog.new("I know this looks bizarre.", "", 1, 1, 0)
					else
						dialog.new(RCMid .. " " .. RCMreturn .. " not supported yet.", "", 1, 1, 0)
					end
					]]
				elseif tonumber(entdetails[2]) == 13 then
					-- Warp token
					if RCMreturn == L.GOTODESTINATION then
						roomx = math.floor(entitydata[tonumber(entdetails[3])].p1 / 40)
						roomy = math.floor(entitydata[tonumber(entdetails[3])].p2 / 30)
						selectedtileset = levelmetadata[(roomy)*20 + (roomx+1)].tileset
						selectedcolor = levelmetadata[(roomy)*20 + (roomx+1)].tilecol
						love.mouse.setPosition(64+64 + (entitydata[tonumber(entdetails[3])].p1 - (roomx*40))*16 + 8, (entitydata[tonumber(entdetails[3])].p2 - (roomy*30))*16 + 8)
						cons("Destination token is at " .. entitydata[tonumber(entdetails[3])].p1 .. " " .. entitydata[tonumber(entdetails[3])].p2 .. "... So at " .. entitydata[tonumber(entdetails[3])].p1 - (roomx*40) .. " " .. entitydata[tonumber(entdetails[3])].p2 - (roomy*30) .. " in room " .. roomx .. " " .. roomy)
					elseif RCMreturn == L.GOTOENTRANCE then
						roomx = math.floor(entitydata[tonumber(entdetails[3])].x / 40)
						roomy = math.floor(entitydata[tonumber(entdetails[3])].y / 30)
						selectedtileset = levelmetadata[(roomy)*20 + (roomx+1)].tileset
						selectedcolor = levelmetadata[(roomy)*20 + (roomx+1)].tilecol
						love.mouse.setPosition(64+64 + (entitydata[tonumber(entdetails[3])].x - (roomx*40))*16 + 8, (entitydata[tonumber(entdetails[3])].y - (roomy*30))*16 + 8)
						cons("Entrance token is at " .. entitydata[tonumber(entdetails[3])].x .. " " .. entitydata[tonumber(entdetails[3])].y .. "... So at " .. entitydata[tonumber(entdetails[3])].x - (roomx*40) .. " " .. entitydata[tonumber(entdetails[3])].y - (roomy*30) .. " in room " .. roomx .. " " .. roomy)
					else
						dialog.new(RCMid .. " " .. RCMreturn .. " not supported yet.", "", 1, 1, 0)
					end
				elseif tonumber(entdetails[2]) == 15 then
					-- Rescuable crewmate
					if RCMreturn == L.CHANGECOLOR then
						entitydata[tonumber(entdetails[3])].p1 = cycle(entitydata[tonumber(entdetails[3])].p1, 5, 0)
					else
						dialog.new(RCMid .. " " .. RCMreturn .. " not supported yet.", "", 1, 1, 0)
					end
				elseif tonumber(entdetails[2]) == 16 then
					-- Start point
					if RCMreturn == L.CHANGEDIRECTION then
						entitydata[tonumber(entdetails[3])].p1 = cycle(entitydata[tonumber(entdetails[3])].p1, 1, 0)
					else
						dialog.new(RCMid .. " " .. RCMreturn .. " not supported yet.", "", 1, 1, 0)
					end
				elseif tonumber(entdetails[2]) == 17 then
					-- Roomtext
					if RCMreturn == L.EDITTEXT then
						-- Were we already editing roomtext or a name?
						if editingroomtext > 0 then
							-- The argument here is the number of the entity not to make nil- the entity we currently want to edit the text of
							endeditingroomtext(tonumber(entdetails[3]))
						end
					
						startinput()
						input = entitydata[tonumber(entdetails[3])].data
						editingroomtext = tonumber(entdetails[3])
						makescriptroomtext = false
					elseif RCMreturn == L.COPYTEXT then
						love.system.setClipboardText(entitydata[tonumber(entdetails[3])].data)
					else
						dialog.new(RCMid .. " " .. RCMreturn .. " not supported yet.", "", 1, 1, 0)
					end
				elseif tonumber(entdetails[2]) == 18 then
					-- Terminal
					if RCMreturn == L.EDITSCRIPT then
						-- Were we editing roomtext or a name?
						if editingroomtext > 0 then
							-- The argument here is the number of the entity not to make nil- the entity we currently want to edit the script of
							endeditingroomtext(tonumber(entdetails[3]))
						end
						
						if scripts[entitydata[tonumber(entdetails[3])].data] == nil then
							dialog.new("Script \"" .. entitydata[tonumber(entdetails[3])].data .. "\" does not exist!", "", 1, 1, 0)
						else
							scriptname = entitydata[tonumber(entdetails[3])].data
							scriptlines = scripts[entitydata[tonumber(entdetails[3])].data]
							processflaglabels()
							tostate(3)
						end
					elseif RCMreturn == L.OTHERSCRIPT then
						-- Were we already editing roomtext or a name?
						if editingroomtext > 0 then
							-- The argument here is the number of the entity not to make nil- the entity we currently want to edit the name of
							endeditingroomtext(tonumber(entdetails[3]))
						end
					
						startinput()
						input = entitydata[tonumber(entdetails[3])].data
						editingroomtext = tonumber(entdetails[3])
						makescriptroomtext = true
					else
						dialog.new(RCMid .. " " .. RCMreturn .. " not supported yet.", "", 1, 1, 0)
					end
				elseif tonumber(entdetails[2]) == 19 then
					-- Script box
					if RCMreturn == L.EDITSCRIPT then
						-- Were we already editing roomtext or a name?
						if editingroomtext > 0 then
							-- The argument here is the number of the entity not to make nil- the entity we currently want to edit the script of
							endeditingroomtext(tonumber(entdetails[3]))
						end
					
						if scripts[entitydata[tonumber(entdetails[3])].data] == nil then
							dialog.new("Script \"" .. entitydata[tonumber(entdetails[3])].data .. "\" does not exist!", "", 1, 1, 0)
						else
							scriptname = entitydata[tonumber(entdetails[3])].data
							scriptlines = scripts[entitydata[tonumber(entdetails[3])].data]
							processflaglabels()
							tostate(3)
						end
					elseif RCMreturn == L.OTHERSCRIPT then
						-- Were we already editing roomtext or a name?
						if editingroomtext > 0 then
							-- The argument here is the number of the entity not to make nil- the entity we currently want to edit the name of
							endeditingroomtext(tonumber(entdetails[3]))
						end
					
						startinput()
						input = entitydata[tonumber(entdetails[3])].data
						editingroomtext = tonumber(entdetails[3])
						makescriptroomtext = true
					else
						dialog.new(RCMid .. " " .. RCMreturn .. " not supported yet.", "", 1, 1, 0)
					end
				else
					dialog.new(langkeys(L.UNKNOWNENTITYTYPE, {anythingbutnil(entdetails[2])}) .. ",\n\nID: " .. RCMid .. "\nReturn value: " .. RCMreturn, "", 1, 1, 0)
				end
			else
				dialog.new("Entity #" .. tonumber(entdetails[3]) .. " no longer exists!", "", 1, 1, 0)
			end
		elseif RCMid == "savemap" then
			if RCMreturn == "Save map" then
				-- You could have just left clicked the button but this is to prevent confusion.
				saveas = (editingmap == "untitled\n" and "untitled" or editingmap) .. "_" .. os.time() .. ".png"
				mapscreenshot:getData():encode("maps/" .. saveas)
				dialog.new("Map image saved as " .. saveas .. "!", "", 1, 1, 0)
			elseif RCMreturn == "Save fullsize map" then
				-- Save a big map!
				if not love.graphics.isSupported("canvas") then
					dialog.new("Sorry, it seems like your graphics card does not support this feature!", "", 1, 1, 0)
				else
					mapcanvas = love.graphics.newCanvas(320*metadata.mapwidth, 240*metadata.mapheight)
					love.graphics.setCanvas(mapcanvas)
					for mry = 0, metadata.mapheight-1 do
						for mrx = 0, metadata.mapwidth-1 do
							displayroom(mrx*0.5*640, mry*0.5*480, roomdata[mry][mrx], levelmetadata[(mry)*20 + (mrx+1)], 0.5) --mapscale
						end
					end
					saveas = (editingmap == "untitled\n" and "untitled" or editingmap) .. "_" .. os.time() .. "_fullsize.png"
					mapcanvas:getImageData():encode("maps/" .. saveas)
					love.graphics.setCanvas()
					mapcanvas = nil
					
					-- Love2d bug regarding canvases and scissors?
					za,zb,zc = love.window.getMode()
					love.window.setMode(za,zb,zc)
					
					collectgarbage("collect")
					dialog.new("Map image saved as " .. saveas .. "!", "", 1, 1, 0)
				end
			else
				dialog.new(RCMid .. " " .. RCMreturn .. " unrecognized.", "", 1, 1, 0)
			end
		else
			dialog.new("Unhandled right click menu!\n\nID: " .. RCMid .. "\nReturn value: " .. RCMreturn, "", 1, 1, 0)
		end
		
		RCMreturn = ""
	end
	
	dialog.update()
end

function love.textinput(char)
	-- Ved should really only accept printable ASCII only when typing...
	if (state == 15) or char:byte(2, 2) == nil then
		if takinginput then
			-- Ugly, but at least won't need another global variable that appears here and there
			if (state == 1) and not nodialog and editingroomname and (char == "e") then
			else
				input = input .. char
			end
			
			cursorflashtime = 0
			
			if state == 3 then
				scriptlines[editingline] = input
			elseif state == 15 and helpeditingline ~= 0 then
				helparticlecontent[helpeditingline] = input
			elseif state == 6 then
				tabselected = 0
			end
		elseif currentmultiinput ~= 0 then
			multiinput[currentmultiinput] = multiinput[currentmultiinput] .. char
		end
	end
end

function love.keypressed(key)
	-- Your privacy is respected.
	keyva.keypressed(key)
	
	-- DEBUG FOR FPS CAP
	if allowdebug and key == "pagedown" then
		fpscap = (fpscap + 1) % 4
	end

	if takinginput then
		if key == "backspace" then
			input = backspace(input)
			
			if state == 3 then
				-- We're checking for the not-yet-changed line, hence why this doesn't cause the exact same backspace problem VVVVVV has
				if scriptlines[editingline] == "" and editingline ~= 1 then
					table.remove(scriptlines, editingline)
					editingline = editingline - 1
					input = anythingbutnil(scriptlines[editingline])
				else
					scriptlines[editingline] = input
				end
			elseif state == 15 then
				-- Exact same story here.
				if helparticlecontent[helpeditingline] == "" and helpeditingline ~= 1 then
					table.remove(helparticlecontent, helpeditingline)
					helpeditingline = helpeditingline - 1
					input = anythingbutnil(helparticlecontent[helpeditingline])
				else
					helparticlecontent[helpeditingline] = input
				end
			elseif state == 6 then
				tabselected = 0
			end
		elseif love.keyboard.isDown("l" .. ctrl) and love.keyboard.isDown("v") then
			input = input .. love.system.getClipboardText()
			
			if state == 3 then
				if input:find("\n") then
					-- CRLF -> LF
					local inputparts = explode("\n", input:gsub("\r\n", "\n"))
					
					scriptlines[editingline] = inputparts[1]
					
					for k = 2, #inputparts do
						table.insert(scriptlines, editingline+(k-1), inputparts[k])
					end
					
					editingline = editingline + #inputparts - 1
					input = anythingbutnil(scriptlines[editingline])
				else
					scriptlines[editingline] = input
				end
			elseif state == 15 then
				-- Same
				if input:find("\n") then
					-- CRLF -> LF
					local inputparts = explode("\n", input:gsub("\r\n", "\n"))
					
					helparticlecontent[helpeditingline] = inputparts[1]
					
					for k = 2, #inputparts do
						table.insert(helparticlecontent, helpeditingline+(k-1), inputparts[k])
					end
					
					helpeditingline = helpeditingline + #inputparts - 1
					input = anythingbutnil(helparticlecontent[helpeditingline])
				else
					helparticlecontent[helpeditingline] = input
				end
			elseif state == 6 then
				tabselected = 0
			end
		elseif key == "left" then
			input, input_r = leftspace(input, input_r)
			
			cursorflashtime = 0

			if state == 3 then
				scriptlines[editingline] = input
			elseif state == 15 and helpeditingline ~= 0 then
				helparticlecontent[helpeditingline] = input
			end
		elseif key == "right" then
			input, input_r = rightspace(input, input_r)
			
			cursorflashtime = 0
			
			if state == 3 then
				scriptlines[editingline] = input
			elseif state == 15 and helpeditingline ~= 0 then
				helparticlecontent[helpeditingline] = input
			end
		elseif key == "delete" then
			_, input_r = rightspace(input, input_r)
		end
	--elseif currentmultiinput ~= 0 then
	elseif #multiinput > 0 then
		if key == "backspace" then
			multiinput[currentmultiinput] = backspace(multiinput[currentmultiinput])
		elseif love.keyboard.isDown("l" .. ctrl) and love.keyboard.isDown("v") then
			multiinput[currentmultiinput] = multiinput[currentmultiinput] .. love.system.getClipboardText()
		elseif love.keyboard.isDown("tab") and love.keyboard.isDown("lshift") then
			if currentmultiinput <= 1 then
				currentmultiinput = #multiinput
			else
				currentmultiinput = currentmultiinput - 1
			end
		elseif love.keyboard.isDown("tab") then
			if currentmultiinput >= #multiinput then
				currentmultiinput = 1
			else
				currentmultiinput = currentmultiinput + 1
			end
		end
	end
	
	if (DIAwindowani ~= 16) and (DIAcanclose == 1) and (key == "return") then
		dialog.push()
		DIAreturn = 1
	elseif (DIAwindowani ~= 16) and (DIAcanclose == 2) and (key == "return") then
		dialog.push()
		DIAreturn = 1
		DIAquitting = 1
	elseif (DIAwindowani ~= 16) and (DIAcanclose == 4) and (key == "return") then
		dialog.push()
		DIAreturn = 2
	elseif (DIAwindowani ~= 16) and (DIAcanclose == 5) and (key == "return") then
		dialog.push()
		DIAreturn = 3
	elseif (state == 0) and (key == "return") then
		stopinput()
		tostate(input)
	--elseif (state == 1) and (key == "z") then -- temporary key, might change
		--zoomscale = 2
	--[[
	elseif nodialog and editingroomtext == 0 and (state == 1) and (key == "s") then -- again, but cycle tileset
		--metadata.tileset = cycle(metadata.tileset, 2)
		if usedtilesets[levelmetadata[(roomy)*20 + (roomx+1)].tileset] == 1 then
			levelmetadata[(roomy)*20 + (roomx+1)].tileset = 1
		else
			levelmetadata[(roomy)*20 + (roomx+1)].tileset = 0
		end
	]]
	--elseif (state == 1) and (key == "t") then -- ^
		--tilespicker = not tilespicker
		
	elseif nodialog and editingroomtext == 0 and (state == 1) and key == "," then
		if love.keyboard.isDown("l" .. ctrl) or love.keyboard.isDown("r" .. ctrl) then
			if selectedtool ~= 14 then
				if selectedsubtool[selectedtool] > 1 then
					selectedsubtool[selectedtool] = selectedsubtool[selectedtool] - 1
				else
					selectedsubtool[selectedtool] = #subtoolimgs[selectedtool]
				end
			end
		else
			if selectedtool > 1 then
				selectedtool = selectedtool - 1
				love.window.setIcon(toolimgicon[selectedtool])
			else
				selectedtool = 17
				love.window.setIcon(toolimgicon[17])
			end
		end
	elseif nodialog and editingroomtext == 0 and (state == 1) and key == "." then
		if love.keyboard.isDown("l" .. ctrl) or love.keyboard.isDown("r" .. ctrl) then
			if selectedtool ~= 14 then
				if selectedsubtool[selectedtool] < #subtoolimgs[selectedtool] then
					selectedsubtool[selectedtool] = selectedsubtool[selectedtool] + 1
				else
					selectedsubtool[selectedtool] = 1
				end
			end
		else
			if selectedtool < 17 then
				selectedtool = selectedtool + 1
				love.window.setIcon(toolimgicon[selectedtool])
			else
				selectedtool = 1
				love.window.setIcon(toolimgicon[1])
			end
		end
		
	elseif nodialog and editingroomtext == 0 and (state == 1) and (key == "p") and (love.keyboard.isDown("l" .. ctrl) or love.keyboard.isDown("r" .. ctrl)) then
		gotostartpointroom()
	elseif nodialog and not editingroomname and editingroomtext == 0 and (state == 1) and (key == "right") then
		-->
		if roomx+1 >= metadata.mapwidth then
			roomx = 0
		else
			roomx = roomx + 1
		end
		
		selectedtileset = levelmetadata[(roomy)*20 + (roomx+1)].tileset
		selectedcolor = levelmetadata[(roomy)*20 + (roomx+1)].tilecol
	elseif nodialog and not editingroomname and editingroomtext == 0 and (state == 1) and (key == "left") then
		--<
		if roomx+1 <= 1 then
			roomx = metadata.mapwidth-1
		else
			roomx = roomx - 1
		end
		
		selectedtileset = levelmetadata[(roomy)*20 + (roomx+1)].tileset
		selectedcolor = levelmetadata[(roomy)*20 + (roomx+1)].tilecol
	elseif nodialog and not editingroomname and editingroomtext == 0 and (state == 1) and (key == "down") then
		--v
		if roomy+1 >= metadata.mapheight then
			roomy = 0
		else
			roomy = roomy + 1
		end
		
		selectedtileset = levelmetadata[(roomy)*20 + (roomx+1)].tileset
		selectedcolor = levelmetadata[(roomy)*20 + (roomx+1)].tilecol
	elseif nodialog and not editingroomname and editingroomtext == 0 and (state == 1) and (key == "up") then
		--^
		if roomy+1 <= 1 then
			roomy = metadata.mapheight-1
		else
			roomy = roomy - 1
		end
		
		selectedtileset = levelmetadata[(roomy)*20 + (roomx+1)].tileset
		selectedcolor = levelmetadata[(roomy)*20 + (roomx+1)].tilecol
	elseif (state == 1) and editingroomname and (key == "return") then
		editingroomname = false
		stopinput()
		levelmetadata[(roomy)*20 + (roomx+1)].roomname = input
	elseif (state == 1) and (editingroomtext > 0) and (key == "return") then
		endeditingroomtext()
	elseif allowdebug and (state == 1) and (key == "\\") and love.keyboard.isDown("lctrl") then
		cons("*** TILESET COLOR CREATOR STARTED FOR TILESET " .. usedtilesets[levelmetadata[(roomy)*20 + (roomx+1)].tileset] .. " ***")
		cons("First select the wall tiles")
		
		tilescreator = true
		cb = {}
		ca = {}
		cs = {}
		creatorstep = 1
		creatorsubstep = 1
		
		selectedtileset = "creator"
		selectedcolor = "creator"
		
		tilesetblocks.creator.tileimg = usedtilesets[levelmetadata[(roomy)*20 + (roomx+1)].tileset]
		
		tilespicker = true
		selectedtool = 1
		
		mousepressed = false
	elseif allowdebug and (state == 1) and (key == "'") then
		-- Just display all tilesets and colors in the console.
		for k,v in pairs(tilesetblocks) do
			cons("==== " .. k .. " ====")
			for k2,v2 in pairs(v.colors) do
				cons("-> " .. k2)
			end
		end
	-- Now come some more of VVVVVV's keybindings!
	elseif nodialog and (state == 1) and (key == "f1") then
		-- Change tileset
		switchtileset()
	elseif nodialog and (state == 1) and (key == "f2") then
		-- Change tilecol
		switchtilecol()
	elseif nodialog and (state == 1) and (key == "f3") then
		-- Change enemy type
		switchenemies()
	elseif nodialog and editingroomtext == 0 and editingroomname == false and (state == 1) and (key == "f4") then
		-- Enemy bounds
		changeenemybounds()
	elseif nodialog and editingroomtext == 0 and editingroomname == false and (state == 1) and (key == "f5") then
		-- Platform bounds
		changeplatformbounds()
	elseif nodialog and (state == 1) and (key == "f10") then
		-- Auto/manual mode
		changedmode()
	elseif nodialog and editingroomtext == 0 and editingroomname == false and (state == 1) and (key == "w") then
		-- Change warp dir
		changewarpdir()
	elseif nodialog and editingroomtext == 0 and editingroomname == false and (state == 1) and (key == "e") then
		-- Edit room name
		toggleeditroomname()
		
		-- Stop that extra e from getting into the roomname...
		nodialog = false
		
		--[[
		if input == "e" then
			input = ""
		else
			input = input:sub(1, -2)
		end
		]]
	elseif nodialog and editingroomtext == 0 and editingroomname == false and (state == 1) and (key == "s") then
		-- Save
		--tostate(8)
		if (love.keyboard.isDown("l" .. ctrl) or love.keyboard.isDown("r" .. ctrl)) and editingmap ~= "untitled\n" then
			-- Quicksave- we have a name already
			
			-- Maybe we have a massive map... Or a slow computer
			temporaryroomname = "Saving..."
			temporaryroomnametimer = 9000 -- Surely saving a map won't take more than ~5 minutes, would it? I mean...
			
			love.graphics.clear(); love.draw(); love.graphics.present()
			
			-- Save now
			savedsuccess, savederror = savelevel(editingmap .. ".vvvvvv", metadata, roomdata, entitydata, levelmetadata, scripts, vedmetadata)
			
			if not savedsuccess then
				-- Why not :c
				dialog.new(L.SAVENOSUCCESS .. anythingbutnil(savederror), "", 1, 1, 0)
			else
				temporaryroomname = "Saved level as " .. editingmap .. ".vvvvvv"
				temporaryroomnametimer = 90
			end
		else
			startmultiinput({(editingmap ~= "untitled\n" and editingmap or "")})
			dialog.new(L.ENTERNAMESAVE .. "\n", "", 1, 4, 10)
		end
	elseif nodialog and editingroomtext == 0 and editingroomname == false and (state == 1) and (key == "l") then
		-- Load
		dialog.new(L.SURELOADLEVEL .. "\n\n(dialog will be save/don't save/cancel later)", "", 1, 3, 3)
	elseif (not holdingzvx) and nodialog and (editingroomtext == 0) and (editingroomname == false) and (state == 1) and ((key == "c") or (key == "v") or (key == "z") or (key == "x") or (key == "y")) then -- Tried cleaning this bit up, later I realized why it was like this
		if (love.keyboard.isDown("l" .. ctrl) or love.keyboard.isDown("r" .. ctrl)) and love.keyboard.isDown("z") then
			-- We goofed, undo.
			if #undobuffer >= 1 then
				if undobuffer[#undobuffer].undotype == "tiles" then
					roomx = undobuffer[#undobuffer].rx
					roomy = undobuffer[#undobuffer].ry
					
					selectedtileset = levelmetadata[(roomy)*20 + (roomx+1)].tileset
					selectedcolor = levelmetadata[(roomy)*20 + (roomx+1)].tilecol
					
					if (undobuffer[#undobuffer].toundotiles[1] == nil) then
						temporaryroomname = "What are you doing?"
						temporaryroomnametimer = 90
					else
						roomdata[roomy][roomx] = table.copy(undobuffer[#undobuffer].toundotiles)
					end
				else
					temporaryroomname = "Unknown undo type \"" .. undobuffer[#undobuffer].undotype .. "\"!"
					temporaryroomnametimer = 90
				end
				
				table.insert(redobuffer, table.copy(undobuffer[#undobuffer]))
				table.remove(undobuffer, #undobuffer)
			end
		elseif (love.keyboard.isDown("l" .. ctrl) or love.keyboard.isDown("r" .. ctrl)) and love.keyboard.isDown("y") then
			-- No wait redo
			if #redobuffer >= 1 then
				if redobuffer[#redobuffer].undotype == "tiles" then
					roomx = redobuffer[#redobuffer].rx
					roomy = redobuffer[#redobuffer].ry
					
					selectedtileset = levelmetadata[(roomy)*20 + (roomx+1)].tileset
					selectedcolor = levelmetadata[(roomy)*20 + (roomx+1)].tilecol
										
					if (redobuffer[#redobuffer].toredotiles[1] == nil) then
						temporaryroomname = "What are you doing?"
						temporaryroomnametimer = 90
					else
						roomdata[roomy][roomx] = table.copy(redobuffer[#redobuffer].toredotiles)
					end
				else
					temporaryroomname = "Unknown undo type \"" .. redobuffer[#redobuffer].undotype .. "\"!"
					temporaryroomnametimer = 90
				end
				
				table.insert(undobuffer, table.copy(redobuffer[#redobuffer]))
				table.remove(redobuffer, #redobuffer)
			end
		elseif (love.keyboard.isDown("l" .. ctrl) or love.keyboard.isDown("r" .. ctrl)) and love.keyboard.isDown("x") then
			-- Cut the room!
			love.system.setClipboardText(getroomcopydata(roomx, roomy))
			
			-- That's only a copy, now reset the room except for the tileset/col
			setroomfromcopy(levelmetadata[roomy*20 + (roomx+1)].tileset .. "," .. levelmetadata[roomy*20 + (roomx+1)].tilecol .. ",0,0,320,240,4,0,0,320,240,0,0,0," .. (",0"):rep(1200), roomx, roomy)
			
			temporaryroomname = "Room cut to clipboard"
			temporaryroomnametimer = 90
		elseif (love.keyboard.isDown("l" .. ctrl) or love.keyboard.isDown("r" .. ctrl)) and love.keyboard.isDown("c") then
			-- Copy the room!
			love.system.setClipboardText(getroomcopydata(roomx, roomy))
			
			temporaryroomname = "Copied room to clipboard"
			temporaryroomnametimer = 90
		elseif (love.keyboard.isDown("l" .. ctrl) or love.keyboard.isDown("r" .. ctrl)) and love.keyboard.isDown("v") then
			-- Try pasting
			setroomfromcopy(love.system.getClipboardText(), roomx, roomy)
			
			if takinginput then
				-- If we're taking input now, then it's gonna be particularly annoying if we paste an entire room in it!!
				cons("TAKING INPUT WHILE PASTING A ROOM")
				stopinput()
			end
		elseif key == "z" then
			-- 3x3 brush
			if selectedtool == 1 or selectedtool == 2 then
				oldzxsubtool = selectedsubtool[selectedtool]
				selectedsubtool[selectedtool] = 2
				
				holdingzvx = true
			end
		elseif key == "x" then
			-- 5x5 brush
			if selectedtool == 1 or selectedtool == 2 then
				oldzxsubtool = selectedsubtool[selectedtool]
				selectedsubtool[selectedtool] = 3
				
				holdingzvx = true
			end
		elseif key == "c" then
			-- Alright, 7x7 brush
			if selectedtool == 1 or selectedtool == 2 then
				oldzxsubtool = selectedsubtool[selectedtool]
				selectedsubtool[selectedtool] = 4
				
				holdingzvx = true
			end
		elseif key == "v" then
			-- And 9x9 brush
			if selectedtool == 1 or selectedtool == 2 then
				oldzxsubtool = selectedsubtool[selectedtool]
				selectedsubtool[selectedtool] = 5
				
				holdingzvx = true
			end
		end
	elseif (state == 3) and (key == "up") and (editingline > 1) then
		scriptlines[editingline] = input .. input_r
		input_r = ""
		__ = "_"
		editingline = editingline - 1
		input = anythingbutnil(scriptlines[editingline])
	elseif (state == 3) and (key == "down") then
		if scriptlines[editingline+1] ~= nil then
			scriptlines[editingline] = input .. input_r
			input_r = ""
			__ = "_"
			editingline = editingline + 1
			input = anythingbutnil(scriptlines[editingline])
		end
	elseif (state == 3) and key == "return" then
		table.insert(scriptlines, editingline+1, "")
		editingline = editingline + 1
		input = anythingbutnil(scriptlines[editingline])
	elseif (state == 6) and key == "return" and tabselected == 0 then
		stopinput()
		
		state6load(input)
	elseif (state == 6) and (((love.keyboard.isDown("lshift") or love.keyboard.isDown("rshift")) and key == "tab") or key == "up") then
		if tabselected > 0 then
			tabselected = tabselected - 1
		end
	elseif (state == 6) and (key == "tab" or key == "down") then --and tabselected < #files then
		tabselected = tabselected + 1
	elseif (state == 6) and key == "escape" then
		tabselected = 0
	elseif (state == 8) and (key == "return") then
		stopinput()
		savedsuccess, savederror = savelevel(input .. ".vvvvvv", metadata, roomdata, entitydata, levelmetadata, scripts, vedmetadata)
		editingmap = input
	elseif state == 10 and (key == "up" or key == "down") then
		if key == "up" then
			scriptlistscroll = scriptlistscroll + 16
			if scriptlistscroll > 0 then
				scriptlistscroll = 0
			end
		elseif key == "down" then
			scriptlistscroll = scriptlistscroll - 16
			-- bounds?
		end
	elseif state == 11 and key == "return" then
		searchscripts, searchrooms, searchroomtext = searchtext(input)
		searchedfor = input
	elseif nodialog and (state == 10 or state == 11 or state == 12) and key == "escape" then
		tostate(1, true)
		nodialog = false
	elseif nodialog and (state == 15) and key == "escape" then
		tostate(oldstate)
		nodialog = false
	elseif nodialog and state == 3 and key == "escape" then
		scriptlines[editingline] = input .. input_r
		scripts[scriptname] = scriptlines
		tostate(10, true)
		nodialog = false
	elseif nodialog and state == 15 and helpeditingline ~= 0 then
		if key == "up" and helpeditingline ~= 1 then
			helparticlecontent[helpeditingline] = input .. input_r
			input_r = ""
			__ = "_"
			helpeditingline = helpeditingline - 1
			input = anythingbutnil(helparticlecontent[helpeditingline])
		elseif key == "down" and helparticlecontent[helpeditingline+1] ~= nil then
			helparticlecontent[helpeditingline] = input .. input_r
			input_r = ""
			__ = "_"
			helpeditingline = helpeditingline + 1
			input = anythingbutnil(helparticlecontent[helpeditingline])
		elseif key == "return" then
			table.insert(helparticlecontent, helpeditingline+1, "")
			helpeditingline = helpeditingline + 1
			input = anythingbutnil(helparticlecontent[helpeditingline])
		end
	elseif allowdebug and (key == "f10") then
		
	elseif allowdebug and (key == "f11") then
		cons("You pressed F11, you get a wall.\n\n***********************************\n* G L O B A L   V A R I A B L E S *\n***********************************\n")
		for k,v in pairs(_G) do
			if type(v) == "boolean" then
				print(k .. " = " .. (v and "true" or "false") .. "\t\t\t[boolean]")
			elseif type(v) == "function" then
				print(k .. "\t\t\t[function]")
			elseif type(v) == "table" then
				print(k .. "\t\t\t[table]")
			elseif type(v) == "userdata" then
				print(k .. "\t\t\t[userdata]")
			else
				print(k .. " = " .. v .. "\t\t\t[" .. type(v) .. "]")
			end
		end
		cons("\n***********************************\n* E N D                           *\n***********************************\n")
	elseif allowdebug and (key == "f12") then
		tostate(0)
	elseif not editingroomname and (editingroomtext == 0) and nodialog and state == 1 then
		for k,v in pairs(toolshortcuts) do
			if key == string.lower(v) then
				selectedtool = k
				love.window.setIcon(toolimgicon[k])
				lefttoolscroll = math.max(16-(48*(selectedtool-1)), -368)
			end
		end
	end
end

function love.keyreleased(key)
	if holdingzvx and (key == "z" or key == "x" or key == "c" or key == "v") then
		if selectedtool == 1 or selectedtool == 2 then
			selectedsubtool[selectedtool] = oldzxsubtool
		end
		holdingzvx = false
	end
end

function love.mousepressed(x, y, button)
	-- Love 0.10 compatibility
	if type(button) == "number" then
		if button == 1 then
			button = "l"
		elseif button == 2 then
			button = "r"
		elseif button == 3 then
			button = "m"
		end
	end
	
	if s.pscale ~= 1 then
		x, y = x*s.pscale^-1, y*s.pscale^-1
	end
	

	if state == 1 and undosaved == 0 and mouseon(128, 0, 640, 480) then
		table.insert(undobuffer, {undotype = "tiles", rx = roomx, ry = roomy, toundotiles = table.copy(roomdata[roomy][roomx]), toredotiles = {}})
		undosaved = #undobuffer
		redobuffer = {}
		cons("XXXX SAVED BEGIN RESULT FOR UNDO")
	end
	

	if (DIAwindowani ~= 16) and (DIAbar == 1) and (button == "l") and (x >= DIAx) and (x <= DIAx+DIAwidth) and (y >= DIAy-17) and (y <= DIAy) then
		DIAmovingwindow = 1
		DIAmovedfromwx = DIAx
		DIAmovedfromwy = DIAy
		DIAmovedfrommx = x
		DIAmovedfrommy = y
	end
	
	if DIAwindowani ~= 16 then
		if (DIAcanclose == 1) and mousein(DIAx+DIAwidth-51, DIAy+DIAwindowani+DIAheight-26, DIAx+DIAwidth-1, DIAy+DIAwindowani+DIAheight-1) then
			dialog.push()
			DIAreturn = 1
		elseif (DIAcanclose == 2) and mousein(DIAx+DIAwidth-51, DIAy+DIAwindowani+DIAheight-26, DIAx+DIAwidth-1, DIAy+DIAwindowani+DIAheight-1) then
			dialog.push()
			DIAreturn = 1
			DIAquitting = 1
		elseif (DIAcanclose == 3 or DIAcanclose == 4) and mousein(DIAx+DIAwidth-51, DIAy+DIAwindowani+DIAheight-26, DIAx+DIAwidth-1, DIAy+DIAwindowani+DIAheight-1) then
			dialog.push()
			DIAreturn = 1
		elseif (DIAcanclose == 5) and mousein(DIAx+DIAwidth-51, DIAy+DIAwindowani+DIAheight-26, DIAx+DIAwidth-1, DIAy+DIAwindowani+DIAheight-1) then
			DIAreturn = 1
		elseif (DIAcanclose == 3 or DIAcanclose == 4 or DIAcanclose == 5) and mousein(DIAx+DIAwidth-106, DIAy+DIAwindowani+DIAheight-26, DIAx+DIAwidth-56, DIAy+DIAwindowani+DIAheight-1) then
			dialog.push()
			DIAreturn = 2
		elseif (DIAcanclose == 5) and mousein(DIAx+DIAwidth-161, DIAy+DIAwindowani+DIAheight-26, DIAx+DIAwidth-106, DIAy+DIAwindowani+DIAheight-1) then
			dialog.push()
			DIAreturn = 3
		end
		
		return
	end
	
	if state == 1 then
		if x < 64 and not (love.keyboard.isDown("lctrl") or love.keyboard.isDown("lshift") or love.keyboard.isDown("rctrl") or love.keyboard.isDown("rshift")) then
			if button == "wu" then
				lefttoolscroll = lefttoolscroll + 16
				lefttoolscrollbounds()
			elseif button == "wd" then
				lefttoolscroll = lefttoolscroll - 16
				lefttoolscrollbounds()
			end
		--[[
		elseif mouseon(love.graphics.getWidth()-(7*16)-1, love.graphics.getHeight()-16-32-2-12-8, (6*16), 8+4) then -- tileset picker
			if dropdown == 1 then
				dropdown = 0
			else
				--dropdown = 1
				rightclickmenu.create({"#1", "Space St.", "#2", "Outside", "Lab", "Warp Zone", "Ship"}, "tileset")
			end
		]]
		elseif mouseon(love.graphics.getWidth()-(7*16)-1, love.graphics.getHeight()-16-16-2-4-8, (6*16), 8+4) then -- show all tiles
			tilespicker = not tilespicker
			
		elseif nodialog and (love.keyboard.isDown("lctrl") or love.keyboard.isDown("lshift") or love.keyboard.isDown("rctrl") or love.keyboard.isDown("rshift")) and button == flipscrollmore((love.system.getOS() == "OS X" and "wd" or "wu")) and mousein(0, 0, love.graphics.getWidth(), love.graphics.getHeight()) then
			if selectedtool > 1 then
				selectedtool = selectedtool - 1
				love.window.setIcon(toolimgicon[selectedtool])
			else
				selectedtool = 17
				love.window.setIcon(toolimgicon[17])
			end
		elseif nodialog and (love.keyboard.isDown("lctrl") or love.keyboard.isDown("lshift") or love.keyboard.isDown("rctrl") or love.keyboard.isDown("rshift")) and button == flipscrollmore((love.system.getOS() == "OS X" and "wu" or "wd")) and mousein(0, 0, love.graphics.getWidth(), love.graphics.getHeight()) then
			if selectedtool < 17 then
				selectedtool = selectedtool + 1
				love.window.setIcon(toolimgicon[selectedtool])
			else
				selectedtool = 1
				love.window.setIcon(toolimgicon[1])
			end
			
		elseif nodialog and button == flipscrollmore((love.system.getOS() == "OS X" and "wd" or "wu")) and mousein(64, 0, love.graphics.getWidth(), love.graphics.getHeight()) and selectedtool ~= 14 then
			if selectedsubtool[selectedtool] > 1 then
				selectedsubtool[selectedtool] = selectedsubtool[selectedtool] - 1
			else
				selectedsubtool[selectedtool] = #subtoolimgs[selectedtool]
			end
		elseif nodialog and button == flipscrollmore((love.system.getOS() == "OS X" and "wu" or "wd")) and mousein(64, 0, love.graphics.getWidth(), love.graphics.getHeight()) and selectedtool ~= 14 then
			if selectedsubtool[selectedtool] < #subtoolimgs[selectedtool] then
				selectedsubtool[selectedtool] = selectedsubtool[selectedtool] + 1
			else
				selectedsubtool[selectedtool] = 1
			end
		end
	elseif state == 3 then
		if button == "wu" then
			scriptscroll = scriptscroll + 16
			if scriptscroll > 0 then
				scriptscroll = 0
			end
		elseif button == "wd" then
			scriptscroll = scriptscroll - 16
			-- bounds?
		end
	elseif state == 9 and button == "r" then
		rightclickmenu.create({"Delete", "Edit script", "Rename"}, "1")
	elseif state == 10 then
		if button == "wu" then
			scriptlistscroll = scriptlistscroll + 16
			if scriptlistscroll > 0 then
				scriptlistscroll = 0
			end
		elseif button == "wd" then
			scriptlistscroll = scriptlistscroll - 16
			-- bounds?
		end
	elseif state == 15 then
		if x <= 25*8 then
			if button == "wu" then
				helplistscroll = helplistscroll + 16
				if helplistscroll > 0 then
					helplistscroll = 0
				end
			elseif button == "wd" then
				helplistscroll = helplistscroll - 16
				-- bounds?
			end
		elseif x >= 25*8+8 then
			if button == "wu" then
				helparticlescroll = helparticlescroll + 16
				if helparticlescroll > 0 then
					helparticlescroll = 0
				end
			elseif button == "wd" then
				helparticlescroll = helparticlescroll - 16
				-- bounds?
			end
		end
	end
end

function love.mousereleased(x, y, button)
	-- Love 0.10 compatibility
	if type(button) == "number" then
		if button == 1 then
			button = "l"
		elseif button == 2 then
			button = "r"
		elseif button == 3 then
			button = "m"
		end
	end

	if s.pscale ~= 1 then
		x, y = x*s.pscale^-1, y*s.pscale^-1
	end
	
	
	if state == 1 and undosaved ~= 0 and undobuffer[undosaved] ~= nil then
		undobuffer[undosaved].toredotiles = table.copy(roomdata[roomy][roomx])
		undosaved = 0
		cons("XXXX SAVED END RESULT FOR UNDO")
	end

	mousepressed = false
	minsmear = -1; maxsmear = -1
	toout = 0
	
	if DIAmovingwindow == 1 then
		DIAx = DIAmovedfromwx + (x-DIAmovedfrommx)
		DIAy = DIAmovedfromwy + (y-DIAmovedfrommy)
		DIAmovingwindow = 0
	end
end

function love.quit()
	if s.askbeforequit then
		if nodialog and not bypassdialog then
			dialog.new("Are you sure you want to quit? You will lose any unsaved content.", "", 1, 3, 8)
			return true
		elseif not nodialog then
			return true
		end
	end
end

function love.threaderror(thread, errorstr)
	dialog.new("Thread error!\n\n" .. errorstr, "", 1, 1, 0)
end